import logging
import os

import geopandas as gpd
import numpy as np
from shapely.geometry import Polygon

logger = logging.getLogger(__name__)


def _polygons_from_model_ds(model_ds):
    """create polygons of each cell in a model dataset.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data

    Raises
    ------
    ValueError
        for wrong gridtype or inconsistent grid definition.

    Returns
    -------
    polygons : list of shapely Polygons
        list with polygon of each raster cell.
    """

    if model_ds.gridtype == 'structured':
        # check if coördinates are consistent with delr/delc values
        delr_x = np.unique(model_ds.x.values[1:] - model_ds.x.values[:-1])
        delc_y = np.unique(model_ds.y.values[:-1] - model_ds.y.values[1:])
        if not ((delr_x == model_ds.delr) and (delc_y == model_ds.delc)):
            raise ValueError(
                'delr and delc attributes of model_ds inconsistent '
                'with x and y coordinates')

        xmins = model_ds.x - (model_ds.delr * 0.5)
        xmaxs = model_ds.x + (model_ds.delr * 0.5)
        ymins = model_ds.y - (model_ds.delc * 0.5)
        ymaxs = model_ds.y + (model_ds.delc * 0.5)
        polygons = [Polygon([(xmins[i], ymins[j]), (xmins[i], ymaxs[j]),
                             (xmaxs[i], ymaxs[j]), (xmaxs[i], ymins[j])])
                    for i in range(len(xmins)) for j in range(len(ymins))]
    elif model_ds.gridtype == 'vertex':
        polygons = [Polygon(vertices)
                    for vertices in model_ds['vertices'].values]
    else:
        raise ValueError(
            "gridtype must be 'structured' or 'vertex', "
            f"not {model_ds.gridtype}")

    return polygons


def vertex_dataarray_to_gdf(model_ds, data_variables, polygons=None,
                            dealing_with_time='mean'):
    """Convert one or more DataArrays from a vertex model dataset to a
    Geodataframe.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    data_variables : list, tuple or set
        data_variables in model_ds that will be stored as attributes in the
        shapefile.
    polygons : list of shapely Polygons, optional
        geometries used for the GeoDataframe, if None the polygons are created
        from the data variable 'vertices' in model_ds. The default is None.
    dealing_with_time : str, optional
        when there is time variant data in the model dataset this function
        becomes very slow. For now only the time averaged data will be
        saved in the geodataframe. Later this can be extended with multiple
        possibilities. The default is 'mean'.

    Raises
    ------
    ValueError
        for DataArrays with unexpected dimension names.
    NotImplementedError
        for DataArrays with more than 2 dimensions or dealing_with_time!='mean'

    Returns
    -------
    gdf : geopandas.GeoDataframe
        geodataframe of one or more DataArrays.
    """
    assert model_ds.gridtype == 'vertex', f'expected model dataset with gridtype vertex, got {model_ds.gridtype}'

    # create dictionary with column names and values of the geodataframe
    dv_dic = {}
    for da_name in data_variables:
        da = model_ds[da_name]
        no_dims = len(da.dims)
        if no_dims == 1:
            dv_dic[da_name] = da.values
        elif no_dims == 2:
            if da.dims == ('layer', 'cid'):
                for i, da_lay in enumerate(da):
                    dv_dic[f'{da_name}_lay{i}'] = da_lay
            elif 'time' in da.dims:
                da_mean = da.mean(dim='time')
                if dealing_with_time == 'mean':
                    dv_dic[f'{da_name}_mean'] = da_mean.values
                else:
                    raise NotImplementedError(
                        "Can only use the mean of a DataArray with dimension time, use dealing_with_time='mean'")
            else:
                raise ValueError(
                    f"expected dimensions ('layer', 'cid'), got {da.dims}")
        else:
            raise NotImplementedError(
                f'expected one or two dimensions got {no_dims} for data variable {da_name}')

    # create geometries
    if polygons is None:
        polygons = _polygons_from_model_ds(model_ds)

    # construct geodataframe
    gdf = gpd.GeoDataFrame(dv_dic, geometry=polygons)

    return gdf


def struc_dataarray_to_gdf(model_ds, data_variables, polygons=None,
                           dealing_with_time='mean'):
    """Convert one or more DataArrays from a structured model dataset to a
    Geodataframe.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    data_variables : list, tuple or set
        data_variables in model_ds that will be stored as attributes in the
        shapefile.
    polygons : list of shapely Polygons, optional
        geometries used for the GeoDataframe, if None the polygons are created
        from the data variable 'vertices' in model_ds. The default is None.

    Raises
    ------
    ValueError
        for DataArrays with unexpected dimension names.
    NotImplementedError
        for DataArrays with more than 2 dimensions or dealing_with_time!='mean'

    Returns
    -------
    gdf : geopandas.GeoDataframe
        geodataframe of one or more DataArrays.
    """
    assert model_ds.gridtype == 'structured', f'expected model dataset with gridtype vertex, got {model_ds.gridtype}'

    # create dictionary with column names and values of the geodataframe
    dv_dic = {}
    for da_name in data_variables:
        da = model_ds[da_name]
        no_dims = len(da.dims)
        if no_dims == 2:
            dv_dic[da_name] = da.values.flatten('F')
        elif no_dims == 3:
            if da.dims == ('layer', 'y', 'x'):
                for i, da_lay in enumerate(da):
                    dv_dic[f'{da_name}_lay{i}'] = da_lay.values.flatten('F')
            elif 'time' in da.dims:
                da_mean = da.mean(dim='time')
                if dealing_with_time == 'mean':
                    dv_dic[f'{da_name}_mean'] = da_mean.values.flatten('F')
                else:
                    raise NotImplementedError(
                        "Can only use the mean of a DataArray with dimension time, use dealing_with_time='mean'")
            else:
                raise ValueError(
                    f"expected dimensions ('layer', 'y', 'x'), got {da.dims}")
        else:
            raise NotImplementedError(
                f'expected two or three dimensions got {no_dims} for data variable {da_name}')

    # create geometries
    if polygons is None:
        polygons = _polygons_from_model_ds(model_ds)

    # construct geodataframe
    gdf = gpd.GeoDataFrame(dv_dic, geometry=polygons)

    return gdf


def dataarray_to_shapefile(model_ds, data_variables, fname, polygons=None):
    """Save one or more DataArrays from a model dataset as a shapefile.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    data_variables : list, tuple or set
        data_variables in model_ds that will be stored as attributes in the
        shapefile.
    fname : str
        filename of the shapefile.
    polygons : list of shapely Polygons, optional
        geometries used for the GeoDataframe, if None the polygons are created
        from the data variable 'vertices' in model_ds. The default is None.


    Returns
    -------
    None.
    """
    if model_ds.gridtype == 'vertex':
        gdf = vertex_dataarray_to_gdf(model_ds, data_variables,
                                      polygons=polygons)
    else:
        gdf = struc_dataarray_to_gdf(model_ds, data_variables,
                                     polygons=polygons)
    gdf.to_file(fname)


def model_dataset_to_vector_file(model_ds,
                                 gisdir=None,
                                 driver='GPKG',
                                 combine_dic=None,
                                 exclude=('x', 'y', 'time_steps', 'area', 'vertices',
                                          'rch_name')):
    """Save all data variables in a model dataset to multiple shapefiles.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    gisdir : str, optional
        gis directory to save shapefiles, if None a subdirectory 'gis' of the
        model_ws (which is an attribute of model_ds) is used. The default is
        None.
    driver : str, optional
        determines if the data variables are saved as seperate "ESRI Shapefile"
        or a single "GPKG" (geopackage). The default is geopackage.
    combine_dic : dictionary of str:set(), optional
        The items in this dictionary are data variables in model_ds that
        should be combined in one shapefile. The key defines the name of the
        shapefile. If None the default combine_dic is used. The default is
        None.
    exclude : tuple of str, optional
        data variables that are not exported to shapefiles. The default is
        ('x', 'y', 'time_steps', 'area', 'vertices').

    Returns
    -------
    fnames : str or list of str
        filename(s) of exported geopackage or shapefiles.
    """

    # get default combination dictionary
    if combine_dic is None:
        combine_dic = {'idomain': {'first_active_layer', 'idomain'},
                       'topbot': {'top', 'bot', 'thickness'},
                       'sea': {'northsea', 'bathymetry'}}

    # create gis directory and filenames
    if gisdir is None:
        gisdir = os.path.join(model_ds.model_ws, 'gis')
    if not os.path.exists(gisdir):
        os.mkdir(gisdir)

    if driver == 'GPKG':
        fname_gpkg = os.path.join(gisdir, f"{model_ds.model_name}.gpkg")
    elif driver == "ESRI Shapefile":
        fnames = []
    else:
        raise ValueError(f'invalid driver -> {driver}')

    # get all data variables in the model dataset
    da_names = set(model_ds)

    # add data variables with only time variant data to exclude list
    for da_name in da_names:
        if ('time',) == model_ds[da_name].dims:
            exclude += (da_name,)

    # exclude some names from export
    da_names -= set(exclude)

    # create list of polygons
    polygons = _polygons_from_model_ds(model_ds)

    # combine some data variables in one shapefile
    for key, item in combine_dic.items():
        if set(item).issubset(da_names):
            if model_ds.gridtype == 'structured':
                gdf = struc_dataarray_to_gdf(model_ds, item, polygons=polygons)
            elif model_ds.gridtype == 'vertex':
                gdf = vertex_dataarray_to_gdf(
                    model_ds, item, polygons=polygons)
            if driver == 'GPKG':
                gdf.to_file(fname_gpkg, layer=key, driver=driver)
            else:
                fname = os.path.join(gisdir, f'{key}.shp')
                gdf.to_file(fname, driver=driver)
                fnames.append(fname)
            da_names -= item
        else:
            logger.info(
                f'could not add {item} into to geopackage because 1 or more of the data variables do not exist')

    # create unique shapefiles for the other data variables
    for da_name in da_names:
        if model_ds.gridtype == 'structured':
            gdf = struc_dataarray_to_gdf(
                model_ds, (da_name,), polygons=polygons)
        elif model_ds.gridtype == 'vertex':
            gdf = vertex_dataarray_to_gdf(
                model_ds, (da_name,), polygons=polygons)
        if driver == 'GPKG':
            gdf.to_file(fname_gpkg, layer=da_name, driver=driver)
        else:
            fname = os.path.join(gisdir, f'{da_name}.shp')
            gdf.to_file(fname, driver=driver)
            fnames.append(fname)

    # return filename(s)
    if driver == 'GPKG':
        return fname_gpkg
    else:
        return fnames



import datetime as dt
import logging
import numbers

import hydropandas as hpd
import numpy as np
import pandas as pd
import xarray as xr

from .. import cache, util

logger = logging.getLogger(__name__)


@cache.cache_netcdf
def get_recharge(model_ds,
                 nodata=None):
    """add multiple recharge packages to the groundwater flow model with knmi
    data by following these steps:

       1. check for each cell (structured or vertex) which knmi measurement
          stations (prec and evap) are the closest.
       2. download precipitation and evaporation data for all knmi stations that
          were found at 1
       3. create a recharge package in which each cell has a reference to a
          timeseries. Timeseries are created for each unique combination of
          precipitation and evaporation. The following packages are created:
            a. the rch package itself in which cells with the same
               precipitation and evaporation stations are defined. This
               package also refers to all the time series package (see b).
            b. the time series packages in which the recharge flux is defined
               for the time steps of the model. Each package contains the
               time series for one or more cels (defined in a).

    Supports structured and unstructred datasets.

    Parameters
    ----------
    model_ds : xr.DataSet
        dataset containing relevant model grid information
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers. If nodata is None the
        nodata value in model_ds is used.
        the default is None.

    Returns
    -------
    model_ds : xr.DataSet
        dataset with spatial model data including the rch raster
    """
    if nodata is None:
        nodata = model_ds.nodata

    start = pd.Timestamp(model_ds.time.data[0])
    #end_ts = pd.Timestamp(model_ds.time.data[-1])
    perlen = model_ds.perlen

    if model_ds.steady_state:
        end = start + pd.Timedelta(model_ds.perlen, unit=model_ds.time_units)
    else:
        if isinstance(perlen, numbers.Number):
            end = pd.Timestamp(model_ds.time.data[-1] + pd.to_timedelta(perlen,
                                                                        unit=model_ds.time_units))
        elif isinstance(perlen, (list, tuple, np.ndarray)):
            end = pd.Timestamp(model_ds.time.data[-1] + pd.to_timedelta(perlen[-1],
                                                                        unit=model_ds.time_units))
        else:
            raise ValueError(f'did not recognise perlen data type {perlen}')

        # include the end day in the time series.
        end = end + pd.Timedelta(1, 'D')

    model_ds_out = util.get_model_ds_empty(model_ds)

    # get recharge data array
    if (model_ds.gridtype == 'structured') and model_ds.steady_state:
        empty_time_array = np.zeros((model_ds.dims['y'],
                                     model_ds.dims['x']))
        model_ds_out['recharge'] = xr.DataArray(empty_time_array,
                                                dims=('y', 'x'),
                                                coords={'x': model_ds.x,
                                                        'y': model_ds.y})
    elif (model_ds.gridtype == 'structured') and (not model_ds.steady_state):
        empty_time_array = np.zeros((model_ds.dims['y'],
                                     model_ds.dims['x'],
                                     model_ds.dims['time']))
        model_ds_out['recharge'] = xr.DataArray(empty_time_array,
                                                dims=('y', 'x', 'time'),
                                                coords={'time': model_ds.time,
                                                        'x': model_ds.x,
                                                        'y': model_ds.y})
    elif (model_ds.gridtype == 'vertex') and model_ds.steady_state:
        empty_time_array = np.zeros((model_ds.dims['cid']))
        model_ds_out['recharge'] = xr.DataArray(empty_time_array,
                                                dims=('cid'),
                                                coords={'cid': model_ds.cid})
    elif (model_ds.gridtype == 'vertex') and (not model_ds.steady_state):
        empty_time_array = np.zeros((model_ds.dims['cid'],
                                     model_ds.dims['time']))
        model_ds_out['recharge'] = xr.DataArray(empty_time_array,
                                                dims=('cid', 'time'),
                                                coords={'time': model_ds.time,
                                                        'cid': model_ds.cid})

    locations, oc_knmi_prec, oc_knmi_evap = get_knmi_at_locations(model_ds,
                                                                  start=start,
                                                                  end=end,
                                                                  nodata=nodata)

    # add closest precipitation and evaporation measurement station to each cell
    locations[['prec_point', 'distance']
              ] = locations.geo.get_nearest_point(oc_knmi_prec)
    locations[['evap_point', 'distance']
              ] = locations.geo.get_nearest_point(oc_knmi_evap)

    # find unique combination of precipitation and evaporation station
    unique_combinations = locations.drop_duplicates(['prec_point', 'evap_point'])[
        ['prec_point', 'evap_point']].values

    for prec_evap in unique_combinations:
        # get locations with the same prec and evap station
        loc_sel = locations.loc[(locations['prec_point'] == prec_evap[0]) & (
            locations['evap_point'] == prec_evap[1])]

        # calculate recharge time series
        prec = oc_knmi_prec.loc[prec_evap[0], 'obs']['RD']
        evap = oc_knmi_evap.loc[prec_evap[1], 'obs']['EV24']
        recharge_ts = (prec - evap).dropna()

        if recharge_ts.index[-1] < (end - pd.Timedelta(1, unit='D')):
            raise ValueError(
                f'no recharge available at precipitation stations {prec_evap[0]} and evaporation station {prec_evap[1]} for date {end}')

        # fill recharge data array
        if model_ds.steady_state:
            rch_average = recharge_ts.mean()
            if model_ds.gridtype == 'structured':
                # add data to model_ds_out
                for row, col in zip(loc_sel.row, loc_sel.col):
                    model_ds_out['recharge'].data[row, col] = rch_average
            elif model_ds.gridtype == 'vertex':
                # add data to model_ds_out
                model_ds_out['recharge'].loc[loc_sel.index] = rch_average
        else:
            model_recharge = pd.Series(index=model_ds.time.data, dtype=float)
            for j, ts in enumerate(model_recharge.index):
                if j < (len(model_recharge) - 1):
                    model_recharge.loc[ts] = recharge_ts.loc[ts:
                                                             model_recharge.index[j + 1]].iloc[:-1].mean()
                else:
                    model_recharge.loc[ts] = recharge_ts.loc[ts:end].iloc[:-1].mean()

            # add data to model_ds_out
            if model_ds.gridtype == 'structured':
                for row, col in zip(loc_sel.row, loc_sel.col):
                    model_ds_out['recharge'].data[row,
                                                  col, :] = model_recharge.values

            elif model_ds.gridtype == 'vertex':
                model_ds_out['recharge'].loc[loc_sel.index,
                                             :] = model_recharge.values

    for datavar in model_ds_out:
        model_ds_out[datavar].attrs['source'] = 'KNMI'
        model_ds_out[datavar].attrs['date'] = dt.datetime.now().strftime(
            '%Y%m%d')
        model_ds_out[datavar].attrs['units'] = 'm/day'

    return model_ds_out


def get_locations_vertex(model_ds, nodata=-999):
    """get dataframe with the locations of the grid cells of a vertex grid.

    Parameters
    ----------
    model_ds : xr.DataSet
        dataset containing relevant model grid information
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers. If nodata is None the
        nodata value in model_ds is used.
        the default is None

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
        includes the columns: x, y and layer
    """
    # get active locations
    cids_active = np.where(model_ds['first_active_layer'] != nodata)[0]

    # create dataframe from active locations
    x = model_ds['x'].sel(cid=cids_active)
    y = model_ds['y'].sel(cid=cids_active)
    layer = model_ds['first_active_layer'].sel(cid=cids_active)
    locations = pd.DataFrame(index=cids_active,
                             data={'x': x, 'y': y, 'layer': layer})
    locations = hpd.ObsCollection(locations)

    return locations


def get_locations_structured(model_ds, nodata=-999):
    """get dataframe with the locations of the grid cells of a structured grid.

    Parameters
    ----------
    model_ds : xr.DataSet
        dataset containing relevant model grid information
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers. If nodata is None the
        nodata value in model_ds is used.
        the default is None

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
        includes the columns: x, y, row, col and layer
    """

    # store x and y mids in locations of active cells
    rows, columns = np.where(model_ds['first_active_layer'] != nodata)
    x = [model_ds['x'].data[col] for col in columns]
    y = [model_ds['y'].data[row] for row in rows]
    layers = [model_ds['first_active_layer'].data[row, col]
              for row, col in zip(rows, columns)]

    locations = hpd.ObsCollection(pd.DataFrame(data={'x': x,
                                                     'y': y,
                                                     'row': rows,
                                                     'col': columns,
                                                     'layer': layers}))

    return locations


def get_knmi_at_locations(model_ds,
                          start='2010', end=None,
                          nodata=-999):
    """get knmi data at the locations of the active grid cells in model_ds.

    Parameters
    ----------
    model_ds : xr.DataSet
        dataset containing relevant model grid information
    start : str or datetime, optional
        start date of measurements that you want, The default is '2010'.
    end :  str or datetime, optional
        end date of measurements that you want, The default is None.
    nodata : int, optional
        if the first_active_layer data array in model_ds has this value,
        it means this cell is inactive in all layers. If nodata is None the
        nodata value in model_ds is used.
        the default is None

    Raises
    ------
    ValueError
        wrong grid type specified.

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
    oc_knmi_prec : hydropandas.ObsCollection
        ObsCollection with knmi data of the precipitation stations.
    oc_knmi_evap : hydropandas.ObsCollection
        ObsCollection with knmi data of the evaporation stations.
    """
    # get locations
    if model_ds.gridtype == 'structured':
        locations = get_locations_structured(model_ds, nodata=nodata)
    elif model_ds.gridtype == 'vertex':
        locations = get_locations_vertex(model_ds, nodata=nodata)
    else:
        raise ValueError('gridtype should be structured or vertex')

    # get knmi data stations closest to any grid cell
    oc_knmi_prec = hpd.ObsCollection.from_knmi(locations=locations,
                                               start=[start],
                                               end=[end],
                                               meteo_vars=["RD"])

    oc_knmi_evap = hpd.ObsCollection.from_knmi(locations=locations,
                                               start=[start],
                                               end=[end],
                                               meteo_vars=["EV24"])

    return locations, oc_knmi_prec, oc_knmi_evap



# -*- coding: utf-8 -*-
"""function to project regis, or a combination of regis and geotop, data on a
modelgrid."""
import datetime as dt
import logging

import numpy as np
import xarray as xr
from scipy.interpolate import griddata

from .. import cache, mdims
from . import geotop

logger = logging.getLogger(__name__)


@cache.cache_netcdf
def get_combined_layer_models(extent, delr, delc,
                              regis_botm_layer=b'AKc',
                              use_regis=True, use_geotop=True,
                              remove_nan_layers=True):
    """combine layer models into a single layer model.

    Possibilities so far include:
        - use_regis -> full model based on regis
        - use_regis and use_geotop -> holoceen of REGIS is filled with geotop


    Parameters
    ----------
    extent : list, tuple or np.array
        desired model extent (xmin, xmax, ymin, ymax)
    delr : int or float,
        cell size along rows, equal to dx
    delc : int or float,
        cell size along columns, equal to dy
    regis_botm_layer : binary str, optional
        regis layer that is used as the bottom of the model. This layer is
        included in the model. the Default is b'AKc' which is the bottom
        layer of regis. call nlmod.regis.get_layer_names() to get a list of
        regis names.
    use_regis : bool, optional
        True if part of the layer model should be REGIS. The default is True.
    use_geotop : bool, optional
        True if part of the layer model should be geotop. The default is True.
    remove_nan_layers : bool, optional
        if True regis and geotop layers with only nans are removed from the
        model. if False nan layers are kept which might be usefull if you want
        to keep some layers that exist in other models. The default is True.

    Returns
    -------
    combined_ds : xarray dataset
        combination of layer models.

    Raises
    ------
    ValueError
        if an invalid combination of layers is used.
    """

    if use_regis:
        regis_ds = get_regis(extent, delr, delc, regis_botm_layer)
    else:
        raise ValueError('layer models without REGIS not supported')

    if use_geotop:
        geotop_ds = geotop.get_geotop(extent, delr, delc, regis_ds)

    if use_regis and use_geotop:
        regis_geotop_ds = add_geotop_to_regis_hlc(regis_ds, geotop_ds)

        combined_ds = regis_geotop_ds
    elif use_regis:
        combined_ds = regis_ds
    else:
        raise ValueError('combination of model layers not supported')

    if remove_nan_layers:
        nlay, lay_sel = get_non_nan_layers(combined_ds)
        combined_ds = combined_ds.sel(layer=lay_sel)
        logger.info(f'removing {nlay} nan layers from the model')

    return combined_ds


@cache.cache_netcdf
def get_regis(extent, delr, delc, botm_layer=b'AKc'):
    """get a regis dataset projected on the modelgrid.

    Parameters
    ----------
    extent : list, tuple or np.array
        desired model extent (xmin, xmax, ymin, ymax)
    delr : int or float,
        cell size along rows, equal to dx
    delc : int or float,
        cell size along columns, equal to dy
    botm_layer : binary str, optional
        regis layer that is used as the bottom of the model. This layer is
        included in the model. the Default is b'AKc' which is the bottom
        layer of regis. call nlmod.regis.get_layer_names() to get a list of
        regis names.

    Returns
    -------
    regis_ds : xarray dataset
        dataset with regis data projected on the modelgrid.
    """
    # check extent
    extent2, _, _ = fit_extent_to_regis(extent, delr, delc)
    for coord1, coord2 in zip(extent, extent2):
        if coord1 != coord2:
            raise ValueError(
                'extent not fitted to regis please fit to regis first, use the nlmod.regis.fit_extent_to_regis function')

    # get local regis dataset
    regis_url = 'http://www.dinodata.nl:80/opendap/REGIS/REGIS.nc'

    regis_ds_raw = xr.open_dataset(regis_url, decode_times=False)

    # set x and y dimensions to cell center
    regis_ds_raw['x'] = regis_ds_raw.x_bounds.mean('bounds')
    regis_ds_raw['y'] = regis_ds_raw.y_bounds.mean('bounds')

    # slice extent
    regis_ds_raw = regis_ds_raw.sel(x=slice(extent[0], extent[1]),
                                    y=slice(extent[2], extent[3]))

    # slice layers
    if isinstance(botm_layer, str):
        botm_layer = botm_layer.encode('utf-8')

    layer_no = np.where((regis_ds_raw.layer == botm_layer).values)[0][0]
    regis_ds_raw = regis_ds_raw.sel(layer=regis_ds_raw.layer[:layer_no + 1])

    # slice data vars
    regis_ds_raw = regis_ds_raw[['top', 'bottom', 'kD', 'c', 'kh', 'kv']]
    regis_ds_raw = regis_ds_raw.rename_vars({'bottom': 'bot'})

    # rename layers
    regis_ds_raw = regis_ds_raw.rename({'layer': 'layer_old'})
    regis_ds_raw.coords['layer'] = regis_ds_raw.layer_old.astype(
        str)  # could also use assign_coords
    regis_ds_raw2 = regis_ds_raw.swap_dims({'layer_old': 'layer'})

    # convert regis dataset to grid
    logger.info('resample regis data to structured modelgrid')
    regis_ds = mdims.resample_dataset_to_structured_grid(regis_ds_raw2,
                                                         extent,
                                                         delr, delc)
    regis_ds.attrs['extent'] = extent
    regis_ds.attrs['delr'] = delr
    regis_ds.attrs['delc'] = delc
    regis_ds.attrs['gridtype'] = 'structured'

    for datavar in regis_ds:
        regis_ds[datavar].attrs['source'] = 'REGIS'
        regis_ds[datavar].attrs['url'] = regis_url
        regis_ds[datavar].attrs['date'] = dt.datetime.now().strftime('%Y%m%d')
        if datavar in ['top', 'bot']:
            regis_ds[datavar].attrs['units'] = 'mNAP'
        elif datavar in ['kh', 'kv']:
            regis_ds[datavar].attrs['units'] = 'm/day'

    return regis_ds


def add_geotop_to_regis_hlc(regis_ds, geotop_ds,
                            float_correction=0.001):
    """Combine geotop and regis in such a way that the holoceen in Regis is
    replaced by the geo_eenheden of geotop.

    Parameters
    ----------
    regis_ds: xarray.DataSet
        regis dataset
    geotop_ds: xarray.DataSet
        geotop dataset
    float_correction: float
        due to floating point precision some floating point numbers that are
        the same are not recognised as the same. Therefore this correction is
        used.

    Returns
    -------
    regis_geotop_ds: xr.DataSet
        combined dataset
    """
    regis_geotop_ds = xr.Dataset()

    # find holoceen (remove all layers above Holoceen)
    layer_no = np.where((regis_ds.layer == 'HLc').values)[0][0]
    new_layers = np.append(geotop_ds.layer.data,
                           regis_ds.layer.data[layer_no + 1:].astype('<U8')).astype('O')

    top = xr.DataArray(dims=('layer', 'y', 'x'),
                       coords={'y': geotop_ds.y, 'x': geotop_ds.x,
                               'layer': new_layers})
    bot = xr.DataArray(dims=('layer', 'y', 'x'),
                       coords={'y': geotop_ds.y, 'x': geotop_ds.x,
                               'layer': new_layers})
    kh = xr.DataArray(dims=('layer', 'y', 'x'),
                      coords={'y': geotop_ds.y, 'x': geotop_ds.x,
                              'layer': new_layers})
    kv = xr.DataArray(dims=('layer', 'y', 'x'),
                      coords={'y': geotop_ds.y, 'x': geotop_ds.x,
                              'layer': new_layers})

    # haal overlap tussen geotop en regis weg
    logger.info('cut geotop layer based on regis holoceen')
    for lay in range(geotop_ds.dims['layer']):
        # Alle geotop cellen die onder de onderkant van het holoceen liggen worden inactief
        mask1 = geotop_ds['top'][lay] <= (
            regis_ds['bot'][layer_no] - float_correction)
        geotop_ds['top'][lay] = xr.where(mask1, np.nan, geotop_ds['top'][lay])
        geotop_ds['bot'][lay] = xr.where(mask1, np.nan, geotop_ds['bot'][lay])
        geotop_ds['kh'][lay] = xr.where(mask1, np.nan, geotop_ds['kh'][lay])
        geotop_ds['kv'][lay] = xr.where(mask1, np.nan, geotop_ds['kv'][lay])

        # Alle geotop cellen waarvan de bodem onder de onderkant van het holoceen ligt, krijgen als bodem de onderkant van het holoceen
        mask2 = geotop_ds['bot'][lay] < regis_ds['bot'][layer_no]
        geotop_ds['bot'][lay] = xr.where(
            mask2 * (~mask1), regis_ds['bot'][layer_no], geotop_ds['bot'][lay])

        # Alle geotop cellen die boven de bovenkant van het holoceen liggen worden inactief
        mask3 = geotop_ds['bot'][lay] >= (
            regis_ds['top'][layer_no] - float_correction)
        geotop_ds['top'][lay] = xr.where(mask3, np.nan, geotop_ds['top'][lay])
        geotop_ds['bot'][lay] = xr.where(mask3, np.nan, geotop_ds['bot'][lay])
        geotop_ds['kh'][lay] = xr.where(mask3, np.nan, geotop_ds['kh'][lay])
        geotop_ds['kv'][lay] = xr.where(mask3, np.nan, geotop_ds['kv'][lay])

        # Alle geotop cellen waarvan de top boven de top van het holoceen ligt, krijgen als top het holoceen van regis
        mask4 = geotop_ds['top'][lay] >= regis_ds['top'][layer_no]
        geotop_ds['top'][lay] = xr.where(
            mask4 * (~mask3), regis_ds['top'][layer_no], geotop_ds['top'][lay])

        # overal waar holoceen inactief is, wordt geotop ook inactief
        mask5 = regis_ds['bot'][layer_no].isnull()
        geotop_ds['top'][lay] = xr.where(mask5, np.nan, geotop_ds['top'][lay])
        geotop_ds['bot'][lay] = xr.where(mask5, np.nan, geotop_ds['bot'][lay])
        geotop_ds['kh'][lay] = xr.where(mask5, np.nan, geotop_ds['kh'][lay])
        geotop_ds['kv'][lay] = xr.where(mask5, np.nan, geotop_ds['kv'][lay])
        if (mask2 * (~mask1)).sum() > 0:
            logger.info(
                f'regis holoceen snijdt door laag {geotop_ds.layer[lay].values}')

    top[:len(geotop_ds.layer), :, :] = geotop_ds['top'].data
    top[len(geotop_ds.layer):, :, :] = regis_ds['top'].data[layer_no + 1:]

    bot[:len(geotop_ds.layer), :, :] = geotop_ds['bot'].data
    bot[len(geotop_ds.layer):, :, :] = regis_ds['bot'].data[layer_no + 1:]

    kh[:len(geotop_ds.layer), :, :] = geotop_ds['kh'].data
    kh[len(geotop_ds.layer):, :, :] = regis_ds['kh'].data[layer_no + 1:]

    kv[:len(geotop_ds.layer), :, :] = geotop_ds['kv'].data
    kv[len(geotop_ds.layer):, :, :] = regis_ds['kv'].data[layer_no + 1:]

    regis_geotop_ds['top'] = top
    regis_geotop_ds['bot'] = bot
    regis_geotop_ds['kh'] = kh
    regis_geotop_ds['kv'] = kv

    _ = [regis_geotop_ds.attrs.update({key: item})
         for key, item in regis_ds.attrs.items()]

    # maak top, bot, kh en kv nan waar de laagdikte 0 is
    mask = (regis_geotop_ds['top'] - regis_geotop_ds['bot']) < float_correction
    for key in ['top', 'bot', 'kh', 'kv']:
        regis_geotop_ds[key] = xr.where(mask, np.nan, regis_geotop_ds[key])
        regis_geotop_ds[key].attrs['source'] = 'REGIS/geotop'
        regis_geotop_ds[key].attrs['regis_url'] = regis_ds[key].url
        regis_geotop_ds[key].attrs['geotop_url'] = geotop_ds[key].url
        regis_geotop_ds[key].attrs['date'] = dt.datetime.now().strftime(
            '%Y%m%d')
        if key in ['top', 'bot']:
            regis_geotop_ds[key].attrs['units'] = 'mNAP'
        elif key in ['kh', 'kv']:
            regis_geotop_ds[key].attrs['units'] = 'm/day'

    return regis_geotop_ds


def fit_extent_to_regis(extent, delr, delc, cs_regis=100.):
    """redifine extent and calculate the number of rows and columns.

    The extent will be redefined so that the borders of the grid (xmin, xmax,
    ymin, ymax) correspond with the borders of the regis grid.

    Parameters
    ----------
    extent : list, tuple or np.array
        original extent (xmin, xmax, ymin, ymax)
    delr : int or float,
        cell size along rows, equal to dx
    delc : int or float,
        cell size along columns, equal to dy
    cs_regis : int or float, optional
        cell size of regis grid. The default is 100..

    Returns
    -------
    extent : list, tuple or np.array
        adjusted extent
    nrow : int
        number of rows.
    ncol : int
        number of columns.
    """
    if isinstance(extent, list):
        extent = extent.copy()
    elif isinstance(extent, (tuple, np.ndarray)):
        extent = list(extent)
    else:
        raise TypeError(
            f'expected extent of type list, tuple or np.ndarray, got {type(extent)}')

    logger.info(f'redefining current extent: {extent}, fit to regis raster')

    for d in [delr, delc]:
        available_cell_sizes = [10., 20., 25.,
                                50., 100., 200., 400., 500., 800.]
        if float(d) not in available_cell_sizes:
            raise NotImplementedError('only this cell sizes can be used for '
                                      f'now -> {available_cell_sizes}')

    # if xmin ends with 100 do nothing, otherwise fit xmin to regis cell border
    if extent[0] % cs_regis != 0:
        extent[0] -= extent[0] % cs_regis

    # get number of columns
    ncol = int(np.ceil((extent[1] - extent[0]) / delr))
    extent[1] = extent[0] + (ncol * delr)  # round xmax up to close grid

    # if ymin ends with 100 do nothing, otherwise fit ymin to regis cell border
    if extent[2] % cs_regis != 0:
        extent[2] -= extent[2] % cs_regis

    nrow = int(np.ceil((extent[3] - extent[2]) / delc))  # get number of rows
    extent[3] = extent[2] + (nrow * delc)  # round ymax up to close grid

    logger.info(
        f'new extent is {extent} model has {nrow} rows and {ncol} columns')

    return extent, nrow, ncol


def get_non_nan_layers(raw_layer_mod, data_var='bot'):
    """get number and name of layers based on the number of non-nan layers.

    Parameters
    ----------
    raw_layer_mod : xarray.Dataset
        dataset with raw layer model from regis or geotop.
    data_var : str
        data var that is used to check if layer mod contains nan values

    Returns
    -------
    nlay : int
        number of active layers within regis_ds_raw.
    lay_sel : list of str
        names of the active layers.
    """
    logger.info('find active layers in raw layer model')

    bot_raw_all = raw_layer_mod[data_var]
    lay_sel = []
    for lay in bot_raw_all.layer.data:
        if not bot_raw_all.sel(layer=lay).isnull().all():
            lay_sel.append(lay)
    nlay = len(lay_sel)

    logger.info(f'there are {nlay} active layers within the extent')

    return nlay, lay_sel


def get_layer_names():
    """get all the available regis layer names.

    Returns
    -------
    layer_names : np.array
        array with names of all the regis layers.
    """

    regis_url = 'http://www.dinodata.nl:80/opendap/REGIS/REGIS.nc'
    layer_names = xr.open_dataset(regis_url).layer.values

    return layer_names


def extrapolate_regis(regis_ds):
    """Fill missing data in layermodel based on nearest interpolation.

    Used for ensuring layer model contains data everywhere. Useful for
    filling in data beneath the sea for coastal groundwater models.

    Parameters
    ----------
    regis_ds : xarray.DataSet
        REGIS DataSet

    Returns
    -------
    regis_ds : xarray.DataSet
        filled REGIS layermodel with nearest interpolation
    """
    # fill layermodel with nearest interpolation (usually for filling in data
    # under the North Sea)
    mask = np.isnan(regis_ds['top']).all('layer')
    if not np.any(mask):
        # all of the model are is inside
        logger.info("No missing data to extrapolate")
        return regis_ds
    x, y = np.meshgrid(regis_ds.x, regis_ds.y)
    points = (x[~mask], y[~mask])
    xi = (x[mask], y[mask])
    for key in list(regis_ds.keys()):
        data = regis_ds[key].data
        for lay in range(len(regis_ds.layer)):
            values = data[lay][~mask]
            data[lay][mask] = griddata(points, values, xi, method='nearest')
    return regis_ds



# -*- coding: utf-8 -*-
"""Created on Fri Dec  3 11:58:49 2021.

@author: oebbe
"""

import numpy as np
import xarray as xr

from .. import cache, util


@cache.cache_netcdf
def get_chd_at_model_edge(model_ds, idomain):
    """get data array which is 1 at every active cell (defined by idomain) at
    the boundaries of the model (xmin, xmax, ymin, ymax). Other cells are 0.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data.
    idomain : xarray.DataArray
        idomain used to get active cells and shape of DataArray

    Returns
    -------
    model_ds_out : xarray.Dataset
        dataset with chd array
    """
    # add constant head cells at model boundaries

    # get mask with grid edges
    xmin = model_ds['x'] == model_ds['x'].min()
    xmax = model_ds['x'] == model_ds['x'].max()
    ymin = model_ds['y'] == model_ds['y'].min()
    ymax = model_ds['y'] == model_ds['y'].max()

    model_ds_out = util.get_model_ds_empty(model_ds)

    if model_ds.gridtype == 'structured':
        mask2d = (ymin | ymax | xmin | xmax)

        # assign 1 to cells that are on the edge and have an active idomain
        model_ds_out['chd'] = xr.zeros_like(idomain)
        for lay in model_ds.layer:
            model_ds_out['chd'].loc[lay] = np.where(
                mask2d & (idomain.loc[lay] == 1), 1, 0)

    elif model_ds.gridtype == 'vertex':
        mask = np.where([xmin | xmax | ymin | ymax])[1]

        # assign 1 to cells that are on the edge, have an active idomain
        model_ds_out['chd'] = xr.zeros_like(idomain)
        model_ds_out['chd'].loc[:, mask] = 1
        model_ds_out['chd'] = xr.where(idomain == 1,
                                       model_ds_out['chd'], 0)

    return model_ds_out



import logging
import warnings

import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm

logger = logging.getLogger(__name__)


def aggregate_surface_water(gdf, method, model_ds=None):
    """Aggregate surface water features.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame containing surfacewater polygons per grid cell.
        Must contain columns 'stage' (waterlevel),
        'c0' (bottom resistance), and 'botm' (bottom elevation)
    method : str, optional
        "area_weighted" for area-weighted params,
        "max_area" for max area params
        "de_lange" for De Lange formula for conductance
    model_ds : xarray.DataSet, optional
        DataSet containing model layer information (only required for
        method='de_lange')

    Returns
    -------
    celldata : pd.DataFrame
        DataFrame with aggregated surface water parameters per grid cell
    """

    required_cols = {"stage", "c0", "botm"}
    missing_cols = required_cols.difference(gdf.columns)
    if len(missing_cols) > 0:
        raise ValueError(f"Missing columns in DataFrame: {missing_cols}")

    # Post process intersection result
    gr = gdf.groupby(by="cellid")
    celldata = pd.DataFrame(index=gr.groups.keys())

    for cid, group in tqdm(gr, desc="Aggregate surface water data"):

        stage, cond, rbot = get_surfacewater_params(
            group, method, cid=cid, model_ds=model_ds)

        celldata.loc[cid, "stage"] = stage
        celldata.loc[cid, "cond"] = cond
        celldata.loc[cid, "rbot"] = rbot
        celldata.loc[cid, "area"] = group.area.sum()

    return celldata


def get_surfacewater_params(group, method, cid=None, model_ds=None,
                            delange_params=None):

    if method == "area_weighted":
        # stage
        stage = agg_area_weighted(group, "stage")
        # cond
        c0 = agg_area_weighted(group, "c0")
        cond = group.area.sum() / c0
        # rbot
        rbot = group["botm"].min()

    elif method == "max_area":
        # stage
        stage = agg_max_area(group, "stage")
        # cond
        c0 = agg_max_area(group, "c0")
        cond = group.area.sum() / c0
        # rbot
        rbot = group["botm"].min()

    elif method == "de_lange":

        # get additional requisite parameters
        if delange_params is None:
            delange_params = {}

        # defaults
        c1 = delange_params.pop("c1", 0.0)
        N = delange_params.pop("N", 1e-3)

        # stage
        stage = agg_area_weighted(group, "stage")

        # cond
        c0 = agg_area_weighted(group, "c0")
        _, _, cond = agg_de_lange(group, cid, model_ds, c1=c1, c0=c0, N=N)

        # rbot
        rbot = group["botm"].min()

    else:
        raise ValueError(f"Method '{method}' not recognized!")

    return stage, cond, rbot


def agg_max_area(gdf, col):
    return gdf.loc[gdf.area.idxmax(), col]


def agg_area_weighted(gdf, col):
    nanmask = gdf[col].isna()
    aw = ((gdf.area * gdf[col]).sum(skipna=True) /
          gdf.loc[~nanmask].area.sum())
    return aw


def agg_de_lange(group, cid, model_ds, c1=0.0, c0=1.0, N=1e-3,
                 crad_positive=True):

    (A, laytop, laybot, kh, kv, thickness) = \
        get_subsurface_params_by_cellid(model_ds, cid)

    rbot = group["botm"].min()

    # select active layers
    active = thickness > 0
    laybot = laybot[active]
    kh = kh[active]
    kv = kv[active]
    thickness = thickness[active]

    # layer thickn.
    H0 = laytop - laybot[laybot < rbot][0]
    ilay = 0
    rlay = np.where(laybot < rbot)[0][0]

    # equivalent hydraulic conductivities
    H = thickness[ilay:rlay + 1]
    kv = kv[ilay:rlay + 1]
    kh = kh[ilay:rlay + 1]
    kveq = np.sum(H) / np.sum(H / kv)
    kheq = np.sum(H * kh) / np.sum(H)

    # length
    len_est = estimate_polygon_length(group)
    li = len_est.sum()
    # correction if group contains multiple shapes
    # but covers whole cell
    if group.area.sum() == A:
        li = A / np.max([model_ds.delr, model_ds.delc])

    # width
    B = group.area.sum(skipna=True) / li

    # mean water level
    p = group.loc[group.area.idxmax(), "stage"]  # waterlevel

    # calculate params
    pstar, cstar, cond = de_lange_eqns(
        A, H0, kveq, kheq, c1, li, B, c0, p, N, crad_positive=crad_positive)

    return pstar, cstar, cond


def get_subsurface_params_by_cellid(model_ds, cid):
    r, c = cid
    A = model_ds.delr * model_ds.delc  # cell area
    laytop = model_ds['top'].isel(x=c, y=r).data
    laybot = model_ds['bot'].isel(x=c, y=r).data
    kv = model_ds['kv'].isel(x=c, y=r).data
    kh = model_ds['kh'].isel(x=c, y=r).data
    thickness = model_ds["thickness"].isel(x=c, y=r).data
    return A, laytop, laybot, kh, kv, thickness


def de_lange_eqns(A, H0, kv, kh, c1, li, Bin, c0, p, N, crad_positive=True):
    """Calculates the conductance according to De Lange.

    Parameters
    ----------
    A : float
        celoppervlak (m2)
    H0 : float
        doorstroomde dikte (m)
    kv : float
        verticale doorlotendheid (m/d)
    kh : float
        horizontale doorlatendheid (m/d)
    c1 : float
        deklaagweerstand (d)
    li : float
        lengte van de waterlopen (m)
    Bin : float
        bodembreedte (m)
    c0 : float
        slootbodemweerstand (d)
    p : float
        water peil
    N : float
        grondwateraanvulling
    crad_positive: bool, optional
        whether to allow negative crad values. If True, crad will be set to 0
        if it is negative.

    Returns
    -------
    float
        Conductance (m2/d)
    """
    if li > 1e-3 and Bin > 1e-3 and A > 1e-3:
        Bcor = max(Bin, 1e-3)  # has no effect
        L = A / li - Bcor
        y = c1 + H0 / kv

        labdaL = np.sqrt(y * kh * H0)
        if L > 1e-3:
            xL = L / (2 * labdaL)
            FL = xL * coth(xL)
        else:
            FL = 0.0

        labdaB = np.sqrt(y * kh * H0 * c0 / (y + c0))
        xB = Bcor / (2 * labdaB)
        FB = xB * coth(xB)

        CL = (c0 + y) * FL + (c0 * L / Bcor) * FB
        if CL == 0.0:
            CB = 1.0
        else:
            CB = (c1 + c0 + H0 / kv) / (CL - c0 * L / Bcor) * CL

        # volgens Kees Maas mag deze ook < 0 zijn...
        # er miste ook een correctie in de log voor anisotropie
        # Crad = max(0., L / (np.pi * np.sqrt(kv * kh))
        #            * np.log(4 * H0 / (np.pi * Bcor)))
        crad = radial_resistance(L, Bcor, H0, kh, kv)
        if crad_positive:
            crad = max([0.0, crad])

        # Conductance
        pSl = Bcor * li / A
        if pSl >= 1.0 - 1e-10:
            Wp = 1 / (pSl / CB) + crad - c1
        else:
            Wp = 1 / ((1. - pSl) / CL + pSl / CB) + crad - c1
        cond = A / Wp

        # cstar, pstar
        cLstar = CL + crad

        pstar = p + N * (cLstar - y) * (y + c0) * L / (Bcor * cLstar + L * y)
        cstar = cLstar * (c0 + y) * (Bcor + L) / (Bcor * cLstar + L * y)

        return pstar, cstar, cond
    else:
        return 0., 0., 0.


def radial_resistance(L, B, H, kh, kv):
    return (L / (np.pi * np.sqrt(kh * kv)) *
            np.log(4 * H * np.sqrt(kh) / (np.pi * B * np.sqrt(kv))))


def coth(x):
    return 1.0 / np.tanh(x)


def estimate_polygon_length(gdf):
    # estimate length from polygon (for shapefactor > 4)
    shape_factor = gdf.length / np.sqrt(gdf.area)

    len_est1 = (gdf.length - np.sqrt(gdf.length**2 - 16 * gdf.area)) / 4
    len_est2 = (gdf.length + np.sqrt(gdf.length**2 - 16 * gdf.area)) / 4
    len_est = pd.concat([len_est1, len_est2], axis=1).max(axis=1)

    # estimate length from minimum rotated rectangle (for shapefactor < 4)
    min_rect = gdf.geometry.apply(lambda g: g.minimum_rotated_rectangle)
    xy = min_rect.apply(lambda g: np.sqrt(
        (np.array(g.exterior.xy[0]) - np.array(g.exterior.xy[0][0]))**2 +
        (np.array(g.exterior.xy[1]) - np.array(g.exterior.xy[1][0]))**2))
    len_est3 = xy.apply(lambda a: np.partition(a.flatten(), -2)[-2])

    # update length estimate where shape factor is lower than 4
    len_est.loc[shape_factor < 4] = len_est3.loc[shape_factor < 4]

    return len_est


def distribute_cond_over_lays(cond, cellid, rivbot, laytop, laybot,
                              idomain=None, kh=None, stage=None):

    if isinstance(rivbot, (np.ndarray, xr.DataArray)):
        rivbot = float(rivbot[cellid])
    if len(laybot.shape) == 3:
        # the grid is structured grid
        laytop = laytop[cellid[0], cellid[1]]
        laybot = laybot[:, cellid[0], cellid[1]]
        if idomain is not None:
            idomain = idomain[:, cellid[0], cellid[1]]
        if kh is not None:
            kh = kh[:, cellid[0], cellid[1]]
    elif len(laybot.shape) == 2:
        # the grid is a vertex grid
        laytop = laytop[cellid]
        laybot = laybot[:, cellid]
        if idomain is not None:
            idomain = idomain[:, cellid]
        if kh is not None:
            kh = kh[:, cellid]

    if stage is None or isinstance(stage, str):
        lays = np.arange(int(np.sum(rivbot < laybot)) + 1)
    elif np.isfinite(stage):
        lays = np.arange(int(np.sum(stage < laybot)),
                         int(np.sum(rivbot < laybot)) + 1)
    else:
        lays = np.arange(int(np.sum(rivbot < laybot)) + 1)
    if idomain is not None:
        # only distribute conductance over active layers
        lays = lays[idomain.values[lays] > 0]
    topbot = np.hstack((laytop, laybot))
    topbot[topbot < rivbot] = rivbot
    d = -1 * np.diff(topbot)
    if kh is not None:
        kd = kh * d
    else:
        kd = d
    if np.all(kd <= 0):
        # when for some reason the kd is 0 in all layers (for example when the
        # river bottom is above all the layers), add to the first active layer
        if idomain is not None:
            try:
                first_active = np.where(idomain == 1)[0][0]
            except IndexError:
                warnings.warn(f"No active layers in {cellid}, "
                              "returning NaNs.")
                return np.nan, np.nan
        else:
            first_active = 0
        lays = [first_active]
        kd[first_active] = 1.
    conds = cond * kd[lays] / np.sum(kd[lays])
    return np.array(lays), np.array(conds)


def build_spd(celldata, pkg, model_ds):
    """Build stress period data for package (RIV, DRN, GHB).

    Parameters
    ----------
    celldata : geopandas.GeoDataFrame
        GeoDataFrame containing data. Cellid must be the index,
        and must have columns
    pkg : str
        Modflow package: RIV, DRN or GHB
    model_ds : xarray.DataSet
        DataSet containing model layer information

    Returns
    -------
    spd : list
        list containing stress period data:
        - RIV: [(cellid), stage, cond, rbot]
        - DRN: [(cellid), elev, cond]
        - GHB: [(cellid), elev, cond]
    """

    spd = []

    for cellid, row in tqdm(celldata.iterrows(),
                            total=celldata.index.size,
                            desc=f"Building stress period data {pkg}"):

        # check if there is an active layer for this cell
        if model_ds.gridtype == 'vertex':
            if (model_ds["idomain"].sel(cid=cellid) == 0).all():
                continue
        elif model_ds.gridtype == 'structured':
            if (model_ds["idomain"].isel(y=cellid[0], x=cellid[1]) == 0).all():
                continue

        # rbot
        if "rbot" in row.index:
            rbot = row["rbot"]
            if np.isnan(rbot):
                raise ValueError(f"rbot is NaN in cell {cellid}")
        elif pkg == "RIV":
            raise ValueError("Column 'rbot' required for building "
                             "RIV package!")
        else:
            rbot = np.nan

        # stage
        stage = row["stage"]

        if np.isnan(stage):
            raise ValueError(f"stage is NaN in cell {cellid}")

        if (stage < rbot) and np.isfinite(rbot):
            logger.warning(f"WARNING: stage below bottom elevation in {cellid}, "
                           "stage reset to rbot!")
            stage = rbot

        # conductance
        cond = row["cond"]

        # check value
        if np.isnan(cond):
            raise ValueError(f"Conductance is NaN in cell {cellid}. Info: area={row.area:.2f} "
                             f"len={row.len_estimate:.2f}, BL={row['rbot']}")

        if cond < 0:
            raise ValueError(f"Conductance is negative in cell {cellid}. Info: area={row.area:.2f} "
                             f"len={row.len_estimate:.2f}, BL={row['rbot']}")

        # if surface water penetrates multiple layers:
        lays, conds = distribute_cond_over_lays(cond,
                                                cellid,
                                                rbot,
                                                model_ds.top,
                                                model_ds.bot,
                                                model_ds.idomain,
                                                model_ds.kh,
                                                stage)
        if "aux" in row:
            auxlist = [row["aux"]]
        else:
            auxlist = []

        if model_ds.gridtype == 'vertex':
            cellid = (cellid,)

        # write SPD
        for lay, cond in zip(lays, conds):
            cid = (lay,) + cellid
            if pkg == "RIV":
                spd.append([cid, stage, cond, rbot] + auxlist)
            elif pkg in ["DRN", "GHB"]:
                spd.append([cid, stage, cond] + auxlist)

    return spd



import datetime as dt
import logging
import warnings
import os
import re
import sys
from shutil import copyfile

import flopy
import geopandas as gpd
import numpy as np
import requests
import xarray as xr
from shapely.geometry import box

logger = logging.getLogger(__name__)


def write_and_run_model(gwf, model_ds, write_model_ds=True,
                        nb_path=None):
    """write modflow files and run the model.

    2 extra options:
        1. write the model dataset to cache
        2. copy the modelscript (typically a Jupyter Notebook) to the model
           workspace with a timestamp.


    Parameters
    ----------
    gwf : flopy.mf6.ModflowGwf
        groundwater flow model.
    model_ds : xarray.Dataset
        dataset with model data.
    write_model_ds : bool, optional
        if True the model dataset is cached. The default is True.
    nb_path : str or None, optional
        full path of the Jupyter Notebook (.ipynb) with the modelscript. The
        default is None. Preferably this path does not have to be given
        manually but there is currently no good option to obtain the filename
        of a Jupyter Notebook from within the notebook itself.
    """

    if nb_path is not None:
        new_nb_fname = f'{dt.datetime.now().strftime("%Y%m%d")}' + \
            os.path.split(nb_path)[-1]
        dst = os.path.join(model_ds.model_ws, new_nb_fname)
        logger.info(f'write script {new_nb_fname} to model workspace')
        copyfile(nb_path, dst)

    if write_model_ds:
        logger.info('write model dataset to cache')
        model_ds.attrs['model_dataset_written_to_disk_on'] = dt.datetime.now(
        ).strftime("%Y%m%d_%H:%M:%S")
        model_ds.to_netcdf(os.path.join(model_ds.attrs['cachedir'],
                                        'full_model_ds.nc'))

    logger.info('write modflow files to model workspace')
    gwf.simulation.write_simulation()
    model_ds.attrs['model_data_written_to_disk_on'] = dt.datetime.now().strftime(
        "%Y%m%d_%H:%M:%S")

    logger.info('run model')
    assert gwf.simulation.run_simulation()[0], 'Modflow run not succeeded'
    model_ds.attrs['model_ran_on'] = dt.datetime.now().strftime(
        "%Y%m%d_%H:%M:%S")


def get_model_dirs(model_ws):
    """Creates a new model workspace directory, if it does not exists yet.
    Within the model workspace directory a few subdirectories are created (if
    they don't exist yet):

    - figure
    - cache

    Parameters
    ----------
    model_ws : str
        model workspace.

    Returns
    -------
    figdir : str
        figure directory inside model workspace.
    cachedir : str
        cache directory inside model workspace.
    """
    figdir = os.path.join(model_ws, 'figure')
    cachedir = os.path.join(model_ws, 'cache')
    if not os.path.exists(model_ws):
        os.makedirs(model_ws)

    if not os.path.exists(figdir):
        os.mkdir(figdir)

    if not os.path.exists(cachedir):
        os.mkdir(cachedir)

    return figdir, cachedir


def get_model_ds_empty(model_ds):
    """get a copy of a model dataset with only grid and time information.

    Parameters
    ----------
    model_ds : xr.Dataset
        dataset with at least the variables layer, x, y and time

    Returns
    -------
    model_ds_out : xr.Dataset
        dataset with only model grid and time information
    """

    return model_ds[list(model_ds.coords)].copy()


def get_da_from_da_ds(da_ds, dims=('y', 'x'), data=None):
    """get a dataarray from model_ds with certain dimensions.

    Parameters
    ----------
    da_ds : xr.Dataset or xr.DataArray
        Dataset or DataArray with at least the dimensions defined in dims
    dims : tuple of string, optional
        dimensions of the desired data array. Default is 'y', 'x'
    data : None, int, float, array_like, optional
        data to fill data array with. if None DataArray is filled with nans.
        Default is None.

    Returns
    -------
    da : xr.DataArray
        DataArray with coordinates from model_ds
    """
    if not isinstance(dims, tuple):
        raise TypeError('keyword argument dims should be of type tuple not {type(dims)}')

    coords = {dim: da_ds[dim] for dim in dims}
    da = xr.DataArray(data, dims=dims,
                      coords=coords)

    return da


def find_most_recent_file(folder, name, extension='.pklz'):
    """ find the most recent file in a folder. File must startwith name and
    end width extension. If you want to look for the most recent folder use
    extension = ''.

    Parameters
    ----------
    folder : str
        path of folder to look for files
    name : str
        find only files that start with this name
    extension : str
        find only files with this extension

    Returns
    -------
    newest_file : str
        name of the most recent file
    """

    i = 0
    for file in os.listdir(folder):
        if file.startswith(name) and file.endswith(extension):
            if i == 0:
                newest_file = os.path.join(folder, file)
                time_prev_file = os.stat(newest_file).st_mtime
            else:
                check_file = os.path.join(folder, file)
                if os.stat(check_file).st_mtime > time_prev_file:
                    newest_file = check_file
                    time_prev_file = os.stat(check_file).st_mtime
            i += 1

    if i == 0:
        return None

    return newest_file


def compare_model_extents(extent1, extent2):
    """check overlap between two model extents.

    Parameters
    ----------
    extent1 : list, tuple or array
        first extent [xmin, xmax, ymin, ymax]
    extent2 : xr.DataSet
        second extent

    Returns
    -------
    int
        several outcomes:
            1: extent1 is completely within extent2
            2: extent2 is completely within extent1
    """

    # option1 extent1 is completely within extent2
    check_xmin = extent1[0] >= extent2[0]
    check_xmax = extent1[1] <= extent2[1]
    check_ymin = extent1[2] >= extent2[2]
    check_ymax = extent1[3] <= extent2[3]

    if check_xmin and check_xmax and check_ymin and check_ymax:
        logger.info('extent1 is completely within extent2 ')
        return 1

    # option2 extent2 is completely within extent1
    if (not check_xmin) and (not check_xmax) and (not check_ymin) and (not check_ymax):
        logger.info('extent2 is completely within extent1')
        return 2

    # option 3 left bound
    if (not check_xmin) and check_xmax and check_ymin and check_ymax:
        logger.info(
            'extent1 is completely within extent2 except for the left bound (xmin)')
        return 3

    # option 4 right bound
    if check_xmin and (not check_xmax) and check_ymin and check_ymax:
        logger.info(
            'extent1 is completely within extent2 except for the right bound (xmax)')
        return 4

    # option 10
    if check_xmin and (not check_xmax) and (not check_ymin) and (not check_ymax):
        logger.info('only the left bound of extent 1 is within extent 2')
        return 10

    raise NotImplementedError('other options are not yet implemented')


def polygon_from_extent(extent):
    """create a shapely polygon from a given extent.

    Parameters
    ----------
    extent : tuple, list or array
        extent (xmin, xmax, ymin, ymax).

    Returns
    -------
    polygon_ext : shapely.geometry.polygon.Polygon
        polygon of the extent.
    """

    bbox = (extent[0], extent[2], extent[1], extent[3])
    polygon_ext = box(*tuple(bbox))

    return polygon_ext


def gdf_from_extent(extent, crs="EPSG:28992"):
    """create a geodataframe with a single polygon with the extent given.

    Parameters
    ----------
    extent : tuple, list or array
        extent.
    crs : str, optional
        coördinate reference system of the extent, default is EPSG:28992
        (RD new)

    Returns
    -------
    gdf_extent : GeoDataFrame
        geodataframe with extent.
    """

    geom_extent = polygon_from_extent(extent)
    gdf_extent = gpd.GeoDataFrame(geometry=[geom_extent],
                                  crs=crs)

    return gdf_extent


def gdf_within_extent(gdf, extent):
    """select only parts of the geodataframe within the extent. Only accepts
    Polygon and Linestring geometry types.

    Parameters
    ----------
    gdf : geopandas GeoDataFrame
        dataframe with polygon features.
    extent : list or tuple
        extent to slice gdf, (xmin, xmax, ymin, ymax).

    Returns
    -------
    gdf : geopandas GeoDataFrame
        dataframe with only polygon features within the extent.
    """
    # create geodataframe from the extent
    gdf_extent = gdf_from_extent(extent, crs=gdf.crs)

    # check type
    geom_types = gdf.geom_type.unique()
    if len(geom_types) > 1:
        # exception if geomtypes is a combination of Polygon and Multipolygon
        multipoly_check = ('Polygon' in geom_types) and (
            'MultiPolygon' in geom_types)
        if (len(geom_types) == 2) and multipoly_check:
            gdf = gpd.overlay(gdf, gdf_extent)
        else:
            raise TypeError(
                f'Only accepts single geometry type not {geom_types}')
    elif geom_types[0] == 'Polygon':
        gdf = gpd.overlay(gdf, gdf_extent)
    elif geom_types[0] == 'LineString':
        gdf = gpd.sjoin(gdf, gdf_extent)
    elif geom_types[0] == 'Point':
        gdf = gdf.loc[gdf.within(gdf_extent.geometry.values[0])]
    else:
        raise TypeError('Function is not tested for geometry type: '
                        f'{geom_types[0]}')

    return gdf


def get_google_drive_filename(fid):
    """get the filename of a google drive file.

    Parameters
    ----------
    fid : str
        google drive id name of a file.

    Returns
    -------
    file_name : str
        filename.
    """
    warnings.warn(
        'this function is no longer supported use the gdown package instead',
        DeprecationWarning)

    if isinstance(id, requests.Response):
        response = id
    else:
        url = 'https://drive.google.com/uc?export=download&id=' + fid
        response = requests.get(url)
    header = response.headers['Content-Disposition']
    file_name = re.search(r'filename="(.*)"', header).group(1)
    return file_name


def download_file_from_google_drive(fid, destination=None):
    """download a file from google drive using it's id.

    Parameters
    ----------
    fid : str
        google drive id name of a file.
    destination : str, optional
        location to save the file to. If destination is None the file is
        written to the current working directory. The default is None.
    """
    warnings.warn(
        'this function is no longer supported use the gdown package instead',
        DeprecationWarning)

    def get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value

        return None

    def save_response_content(response, destination):
        CHUNK_SIZE = 32768

        with open(destination, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={'id': fid}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': fid, 'confirm': token}
        response = session.get(URL, params=params, stream=True)

    if destination is None:
        destination = get_google_drive_filename(fid)
    else:
        if os.path.isdir(destination):
            filename = get_google_drive_filename(fid)
            destination = os.path.join(destination, filename)

    save_response_content(response, destination)


# %% helper functions (from USGS)


def get_platform(pltfrm):
    """Determine the platform in order to construct the zip file name.

    Source: USGS

    Parameters
    ----------
    pltfrm : str, optional
        check if platform string is correct for downloading binaries,
        default is None and will determine platform string based on system

    Returns
    -------
    pltfrm : str
        return platform string
    """
    if pltfrm is None:
        if sys.platform.lower() == 'darwin':
            pltfrm = 'mac'
        elif sys.platform.lower().startswith('linux'):
            pltfrm = 'linux'
        elif 'win' in sys.platform.lower():
            is_64bits = sys.maxsize > 2 ** 32
            if is_64bits:
                pltfrm = 'win64'
            else:
                pltfrm = 'win32'
        else:
            errmsg = ('Could not determine platform'
                      '.  sys.platform is {}'.format(sys.platform))
            raise Exception(errmsg)
    else:
        assert pltfrm in ['mac', 'linux', 'win32', 'win64']
    return pltfrm


def getmfexes(pth='.', version='', pltfrm=None):
    """Get the latest MODFLOW binary executables from a github site
    (https://github.com/MODFLOW-USGS/executables) for the specified operating
    system and put them in the specified path.

    Source: USGS

    Parameters
    ----------
    pth : str
        Location to put the executables (default is current working directory)

    version : str
        Version of the MODFLOW-USGS/executables release to use.

    pltfrm : str
        Platform that will run the executables.  Valid values include mac,
        linux, win32 and win64.  If platform is None, then routine will
        download the latest appropriate zipfile from the github repository
        based on the platform running this script.
    """
    try:
        import pymake
    except ModuleNotFoundError as e:
        print("Install pymake with "
              "`pip install "
              "https://github.com/modflowpy/pymake/zipball/master`")
        raise e
    # Determine the platform in order to construct the zip file name
    pltfrm = get_platform(pltfrm)
    zipname = '{}.zip'.format(pltfrm)

    # Determine path for file download and then download and unzip
    url = ('https://github.com/MODFLOW-USGS/executables/'
           'releases/download/{}/'.format(version))
    assets = {p: url + p for p in ['mac.zip', 'linux.zip',
                                   'win32.zip', 'win64.zip']}
    download_url = assets[zipname]
    pymake.download_and_unzip(download_url, pth)


def get_heads_dataarray(model_ds, fill_nans=False, fname_hds=None):
    """reads the heads from a modflow .hds file and returns an xarray
    DataArray.

    Parameters
    ----------
    model_ds : TYPE
        DESCRIPTION.
    fill_nans : bool, optional
        if True the nan values are filled with the heads in the cells below
    fname_hds : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    head_ar : TYPE
        DESCRIPTION.
    """

    if fname_hds is None:
        fname_hds = os.path.join(
            model_ds.model_ws, model_ds.model_name + '.hds')

    head_filled = get_heads_array(fname_hds, gridtype=model_ds.gridtype,
                                  fill_nans=fill_nans)

    if model_ds.gridtype == 'vertex':
        head_ar = xr.DataArray(data=head_filled[:, :, :],
                               dims=('time', 'layer', 'cid'),
                               coords={'cid': model_ds.cid,
                                       'layer': model_ds.layer,
                                       'time': model_ds.time})
    elif model_ds.gridtype == 'structured':
        head_ar = xr.DataArray(data=head_filled,
                               dims=('time', 'layer', 'y', 'x'),
                               coords={'x': model_ds.x,
                                       'y': model_ds.y,
                                       'layer': model_ds.layer,
                                       'time': model_ds.time})

    return head_ar


def get_heads_array(fname_hds, gridtype='structured',
                    fill_nans=False):
    """reads the heads from a modflow .hds file and returns a numpy array.

    assumes the dimensions of the heads file are:
        structured: time, layer, cid
        vertex: time, layer, nrow, ncol


    Parameters
    ----------
    fname_hds : TYPE, optional
        DESCRIPTION. The default is None.
    gridtype : str, optional
        DESCRIPTION. The default is 'structured'.
    fill_nans : bool, optional
        if True the nan values are filled with the heads in the cells below

    Returns
    -------
    head_ar : np.ndarray
        heads array.
    """
    hdobj = flopy.utils.HeadFile(fname_hds)
    head = hdobj.get_alldata()
    # TODO: this will sometimes set largest head to NaN...
    head[head == head.max()] = np.nan

    if gridtype == 'vertex':
        head_filled = np.ones(
            (head.shape[0], head.shape[1], head.shape[3])) * np.nan

        for t in range(head.shape[0]):
            for lay in range(head.shape[1] - 1, -1, -1):
                head_filled[t][lay] = head[t][lay][0]
                if lay < (head.shape[1] - 1):
                    if fill_nans:
                        head_filled[t][lay] = np.where(np.isnan(head_filled[t][lay]),
                                                       head_filled[t][lay + 1],
                                                       head_filled[t][lay])

    elif gridtype == 'structured':
        head_filled = np.zeros_like(head)
        for t in range(head.shape[0]):
            for lay in range(head.shape[1] - 1, -1, -1):
                head_filled[t][lay] = head[t][lay]
                if lay < (head.shape[1] - 1):
                    if fill_nans:
                        head_filled[t][lay] = np.where(np.isnan(head_filled[t][lay]),
                                                       head_filled[t][lay + 1],
                                                       head_filled[t][lay])
    else:
        raise ValueError('wrong gridtype')

    return head_filled


def download_mfbinaries(binpath=None, version='6.0'):
    """Download and unpack platform-specific modflow binaries.

    Source: USGS

    Parameters
    ----------
    binpath : str, optional
        path to directory to download binaries to, if it doesnt exist it
        is created. Default is None which sets dir to nlmod/bin.
    version : str, optional
        version string, by default '6.0'
    """
    if binpath is None:
        binpath = os.path.join(os.path.dirname(__file__), "..", "bin")
    pltfrm = get_platform(None)
    # Download and unpack mf6 exes
    getmfexes(pth=binpath, version=version, pltfrm=pltfrm)



# -*- coding: utf-8 -*-
"""Created on Fri Jun 12 15:33:03 2020.

@author: ruben
"""

import datetime as dt
import logging
import tempfile

import numpy as np
import xarray as xr
from owslib.wcs import WebCoverageService
from rasterio import merge
from rasterio.io import MemoryFile

from .. import cache, mdims, util

logger = logging.getLogger(__name__)


@cache.cache_netcdf
def get_ahn(model_ds, identifier='ahn3_5m_dtm', gridprops=None):
    """Get a model dataset with ahn variable.

    Parameters
    ----------
    model_ds : xr.Dataset
        dataset with the model information.
    identifier : str, optional
        Possible values for identifier are:
            'ahn2_05m_int'
            'ahn2_05m_non'
            'ahn2_05m_ruw'
            'ahn2_5m'
            'ahn3_05m_dsm'
            'ahn3_05m_dtm'
            'ahn3_5m_dsm'
            'ahn3_5m_dtm'

        The default is 'ahn3_5m_dtm'.
    gridprops : dictionary, optional
        dictionary with grid properties output from gridgen. Only used if
        gridtype = 'vertex'

    Returns
    -------
    model_ds_out : xr.Dataset
        dataset with the ahn variable.
    """
    if (model_ds.gridtype == 'vertex') and (gridprops is None):
        raise ValueError(
            'gridprops should be specified when gridtype is vertex')

    url = _infer_url(identifier)

    ahn_ds_raw = get_ahn_within_extent(extent=model_ds.extent,
                                       url=url,
                                       identifier=identifier)

    ahn_ds_raw = xr.open_rasterio(ahn_ds_raw.open())
    ahn_ds_raw = ahn_ds_raw.rename({'band': 'layer'})
    nodata = ahn_ds_raw.attrs['nodatavals'][0]
    ahn_ds_raw = ahn_ds_raw.where(ahn_ds_raw != nodata)

    if model_ds.gridtype == 'structured':
        ahn_ds = mdims.resample_dataarray3d_to_structured_grid(ahn_ds_raw,
                                                               extent=model_ds.extent,
                                                               delr=model_ds.delr,
                                                               delc=model_ds.delc,
                                                               xmid=model_ds.x.data,
                                                               ymid=model_ds.y.data)
    elif model_ds.gridtype == 'vertex':
        xyi, cid = mdims.get_xyi_cid(gridprops)
        ahn_ds = mdims.resample_dataarray3d_to_vertex_grid(ahn_ds_raw,
                                                           gridprops,
                                                           xyi, cid)

    model_ds_out = util.get_model_ds_empty(model_ds)
    model_ds_out['ahn'] = ahn_ds[0]

    for datavar in model_ds_out:
        model_ds_out[datavar].attrs['source'] = identifier
        model_ds_out[datavar].attrs['url'] = url
        model_ds_out[datavar].attrs['date'] = dt.datetime.now().strftime(
            '%Y%m%d')
        if datavar == 'ahn':
            model_ds_out[datavar].attrs['units'] = 'mNAP'

    return model_ds_out


def split_ahn_extent(extent, res, x_segments, y_segments, maxsize,
                     tmp_dir=None, **kwargs):
    """There is a max height and width limit of 800 * res for the wcs server.
    This function splits your extent in chunks smaller than the limit. It
    returns a list of gdal Datasets.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent
    res : float
        The resolution of the requested output-data
    x_segments : int
        number of tiles on the x axis
    y_segments : int
        number of tiles on the y axis
    maxsize : int or float
        maximum widht or height of ahn tile
    tmp_dir : str, optional
        Path-like to cache the downloads
    **kwargs :
        keyword arguments of the get_ahn_extent function.

    Returns
    -------
    MemoryFile
        Rasterio MemoryFile of the merged AHN

    Notes
    -----
    1. The resolution is used to obtain the ahn from the wcs server. Not sure
    what kind of interpolation is used to resample the original grid.
    """

    # needs a temporary folder to store the individual ahn tiffs before merge
    with tempfile.TemporaryDirectory() as tempfile_tmp_dir:
        if tmp_dir is None:
            logger.info(f"- Created temporary directory {tempfile_tmp_dir}. "
                        "To store ahn tiffs of subextents")
            tmp_dir_path = tempfile_tmp_dir
        else:
            logger.info(f"- Use {tmp_dir} to store ahn tiffs of subextents")
            tmp_dir_path = tmp_dir

        # write tiles
        datasets = []
        start_x = extent[0]
        for tx in range(x_segments):
            if (tx + 1) == x_segments:
                end_x = extent[1]
            else:
                end_x = start_x + maxsize * res
            start_y = extent[2]
            for ty in range(y_segments):
                if (ty + 1) == y_segments:
                    end_y = extent[3]
                else:
                    end_y = start_y + maxsize * res
                subextent = [start_x, end_x, start_y, end_y]
                logger.info(f'downloading subextent {subextent}')
                logger.info(f'x_segment-{tx}, y_segment-{ty}')

                datasets.append(
                    get_ahn_within_extent(
                        subextent, res=res, tmp_dir=tmp_dir_path, **kwargs))
                start_y = end_y

            start_x = end_x

        memfile = MemoryFile()
        merge.merge([b.open() for b in datasets], dst_path=memfile)

    return memfile


def _infer_url(identifier=None):
    """infer the url from the identifier.

    Parameters
    ----------
    identifier : TYPE, optional
        DESCRIPTION. The default is None.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    url : TYPE
        DESCRIPTION.
    """

    # infer url from identifier
    if 'ahn2' in identifier:
        url = ('https://geodata.nationaalgeoregister.nl/ahn2/wcs?'
               'request=GetCapabilities&service=WCS')
    elif 'ahn3' in identifier:
        url = ('https://geodata.nationaalgeoregister.nl/ahn3/wcs?'
               'request=GetCapabilities&service=WCS')
    else:
        ValueError(f'unknown identifier -> {identifier}')

    return url


def get_ahn_within_extent(extent=None, identifier='ahn3_5m_dtm', url=None,
                          res=None, version='1.0.0', fmt='GEOTIFF_FLOAT32',
                          crs='EPSG:28992', maxsize=800, tmp_dir=None):
    """

    Parameters
    ----------
    extent : list, tuple or np.array, optional
        extent. The default is None.
    identifier : str, optional
        Possible values for identifier are:
            'ahn2_05m_int'
            'ahn2_05m_non'
            'ahn2_05m_ruw'
            'ahn2_5m'
            'ahn3_05m_dsm'
            'ahn3_05m_dtm'
            'ahn3_5m_dsm'
            'ahn3_5m_dtm'

        The default is 'ahn3_5m_dtm'.

        the identifier also contains resolution and type info:
        - 5m or 05m is a resolution of 5x5 or 0.5x0.5 meter.
        - 'dtm' is only surface level (maaiveld), 'dsm' has other surfaces
        such as building.
    url : str or None, optional
        possible values None, 'ahn2' and 'ahn3'. If None the url is inferred
        from the identifier. The default is None.
    res : float, optional
        resolution of ahn raster. If None the resolution is inferred from the
        identifier. The default is None.
    version : str, optional
        version of wcs service, options are '1.0.0' and '2.0.1'.
        The default is '1.0.0'.
    fmt : str, optional
        geotif format . The default is 'GEOTIFF_FLOAT32'.
    crs : str, optional
        coördinate reference system. The default is 'EPSG:28992'.
    tmp_dir : str
        Path-like to temporairly store the downloads before merge.
    maxsize : float, optional
        maximum number of cells in x or y direction. The default is
        800.

    Returns
    -------
    MemoryFile
        Rasterio MemoryFile of the AHN

    """

    if isinstance(extent, xr.DataArray):
        extent = tuple(extent.values)

    # get url
    if url is None:
        url = _infer_url(identifier)
    elif url == 'ahn2':
        url = ('https://geodata.nationaalgeoregister.nl/ahn2/wcs?'
               'request=GetCapabilities&service=WCS')
    elif url == 'ahn3':
        url = ('https://geodata.nationaalgeoregister.nl/ahn3/wcs?'
               'request=GetCapabilities&service=WCS')
    elif not url.startswith("https://geodata.nationaalgeoregister.nl"):
        raise ValueError(f'unknown url -> {url}')

    # check resolution
    if res is None:
        if '05m' in identifier.split('_')[1]:
            res = 0.5
        elif '5m' in identifier.split('_')[1]:
            res = 5.0
        else:
            raise ValueError('could not infer resolution from identifier')

    # check if ahn is within limits
    dx = extent[1] - extent[0]
    dy = extent[3] - extent[2]

    # check if size exceeds maxsize
    if (dx / res) > maxsize:
        x_segments = int(np.ceil((dx / res) / maxsize))
    else:
        x_segments = 1

    if (dy / res) > maxsize:
        y_segments = int(np.ceil((dy / res) / maxsize))
    else:
        y_segments = 1

    if (x_segments * y_segments) > 1:
        st = f'''requested ahn raster width or height bigger than {maxsize*res}
            -> splitting extent into {x_segments} * {y_segments} tiles'''
        logger.info(st)
        return split_ahn_extent(extent, res, x_segments, y_segments, maxsize,
                                identifier=identifier,
                                version=version, fmt=fmt, crs=crs,
                                tmp_dir=tmp_dir)

    # download file
    logger.info(f"- download ahn between: x ({str(extent[0])}, {str(extent[1])}); "
                f"y ({str(extent[2])}, {str(extent[3])})")
    wcs = WebCoverageService(url, version=version)
    if version == '1.0.0':
        bbox = (extent[0], extent[2], extent[1], extent[3])
        output = wcs.getCoverage(identifier=identifier, bbox=bbox,
                                 format=fmt, crs=crs, resx=res,
                                 resy=res)
    elif version == '2.0.1':
        # bbox, resx and resy do nothing in version 2.0.1
        subsets = [('x', extent[0], extent[1]),
                   ('y', extent[2], extent[3])]
        output = wcs.getCoverage(identifier=[identifier], subsets=subsets,
                                 format=fmt, crs=crs)
    else:
        raise (Exception('Version {} not yet supported'.format(version)))

    return MemoryFile(output.read())



import flopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Polygon, Point

from .. import mdims


def line2hfb(gdf, gwf, prevent_rings=True, plot=False):
    """ Obtain the cells with a horizontal flow barrier between them from a
    geodataframe with line elements.


    Parameters
    ----------
    gdf : gpd.GeoDataframe
        geodataframe with line elements.
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        grondwater flow model.
    prevent_rings : bool, optional
        DESCRIPTION. The default is True.
    plot : bool, optional
        If True create a simple plot of the grid cells and shapefile. For a
        more complex plot you can use plot_hfb. The default is False.

    Returns
    -------
    cellids : 2d list of ints
        a list with pairs of cells that have a hfb between them.

    """
    # for the idea, sea:
    # https://gis.stackexchange.com/questions/188755/how-to-snap-a-road-network-to-a-hexagonal-grid-in-qgis

    gdfg = mdims.gdf2grid(gdf, gwf)

    cell2d = pd.DataFrame(gwf.disv.cell2d.array).set_index('icell2d')
    vertices = pd.DataFrame(gwf.disv.vertices.array).set_index('iv')

    # for every cell determine which cell-edge could form the line
    # by testing for an intersection with a triangle to the cell-center
    icvert = cell2d.loc[:, cell2d.columns.str.startswith('icvert')].values

    hfb_seg = []
    for index in gdfg.index.unique():
        # Get the nearest hexagon sides where routes cross
        for icell2d in gdfg.loc[index, 'cellid']:
            for i in range(cell2d.at[icell2d, 'ncvert'] - 1):
                iv1 = icvert[icell2d, i]
                iv2 = icvert[icell2d, i + 1]
                # make sure vert1 is lower than vert2
                if iv1 > iv2:
                    iv1, iv2 = iv2, iv1
                coords = [(cell2d.at[icell2d, 'xc'], cell2d.at[icell2d, 'yc']),
                          (vertices.at[iv1, 'xv'], vertices.at[iv1, 'yv']),
                          (vertices.at[iv2, 'xv'], vertices.at[iv2, 'yv'])]
                triangle = Polygon(coords)
                if triangle.intersects(gdf.loc[index, 'geometry']):
                    hfb_seg.append((icell2d, iv1, iv2))
    hfb_seg = np.array(hfb_seg)

    if prevent_rings:
        # find out if there are cells with segments on each side
        # remove the segments whose centroid is farthest from the line
        for icell2d in np.unique(hfb_seg[:, 0]):
            mask = hfb_seg[:, 0] == icell2d
            if mask.sum() >= cell2d.at[icell2d, 'ncvert'] - 1:
                segs = hfb_seg[mask]
                dist = []
                for seg in segs:
                    p = Point(vertices.loc[seg[1:3], 'xv'].mean(),
                              vertices.loc[seg[1:3], 'yv'].mean())
                    dist.append(gdf.distance(p).min())
                iv1, iv2 = segs[np.argmax(dist), [1, 2]]
                mask = (hfb_seg[:, 1] == iv1) & (hfb_seg[:, 2] == iv2)
                hfb_seg = hfb_seg[~mask]

    # get unique segments
    hfb_seg = np.unique(hfb_seg[:, 1:], axis=0)

    # Get rid of disconnected (or 'open') segments
    # Let's remove disconnected/open segments
    iv = np.unique(hfb_seg)
    segments_per_iv = pd.Series([np.sum(hfb_seg == x) for x in iv], index=iv)
    mask = np.full(hfb_seg.shape[0], True)
    for i, segment in enumerate(hfb_seg):
        # one vertex is not connected and the other one at least to two other segments
        if ((segments_per_iv[segment[0]] == 1 and
             segments_per_iv[segment[1]] >= 3) or
            (segments_per_iv[segment[1]] == 1 and
             segments_per_iv[segment[0]] >= 3)):
            mask[i] = False
    hfb_seg = hfb_seg[mask]

    if plot:
        # test by plotting
        ax = gdfg.plot()
        for i, seg in enumerate(hfb_seg):
            x = [vertices.at[seg[0], 'xv'], vertices.at[seg[1], 'xv']]
            y = [vertices.at[seg[0], 'yv'], vertices.at[seg[1], 'yv']]
            ax.plot(x, y)

    # find out between which cellid's these segments are
    segments = []
    for icell2d in cell2d.index:
        for i in range(cell2d.at[icell2d, 'ncvert'] - 1):
            iv1 = icvert[icell2d, i]
            iv2 = icvert[icell2d, i + 1]
            # make sure vert1 is lower than vert2
            if iv1 > iv2:
                iv1, iv2 = iv2, iv1
            segments.append((icell2d, (iv1, iv2)))
    segments = pd.DataFrame(segments, columns=['icell2d', 'verts'])
    segments = segments.set_index(['verts'])

    cellids = []
    for seg in hfb_seg:
        cellids.append(list(segments.loc[[tuple(seg)]].values[:, 0]))
    return cellids


def plot_hfb(cellids, gwf, ax=None):
    """ plots a horizontal flow barrier


    Parameters
    ----------
    cellids : list of lists of integers or flopy.mf6.ModflowGwfhfb
        list with the ids of adjacent cells that should get a horizontal
        flow barrier, hfb is the output of line2hfb.
    gwf : flopy groundwater flow model
        DESCRIPTION.
    ax : matplotlib axes


    Returns
    -------
    fig : TYPE
        DESCRIPTION.
    ax : TYPE
        DESCRIPTION.

    """
    if ax is None:
        _, ax = plt.subplots()

    if isinstance(cellids, flopy.mf6.ModflowGwfhfb):
        spd = cellids.stress_period_data.data[0]
        cellids = [[line[0][1], line[1][1]] for line in spd]

    for line in cellids:
        pc1 = Polygon(gwf.modelgrid.get_cell_vertices(line[0]))
        pc2 = Polygon(gwf.modelgrid.get_cell_vertices(line[1]))
        x, y = pc1.intersection(pc2).xy
        ax.plot(x, y, color='red')

    return ax



# -*- coding: utf-8 -*-
"""module with functions to deal with the northsea by:

    - identifying model cells with the north sea
    - add bathymetry of the northsea to the layer model
    - extrpolate the layer model below the northsea bed.


Note: if you like jazz please check this out: https://www.northseajazz.com
"""
import datetime as dt
import logging
import os

import numpy as np
import requests
import xarray as xr

from .. import cache, mdims, util

logger = logging.getLogger(__name__)


@cache.cache_netcdf
def get_bathymetry(model_ds, northsea,
                   gridprops=None):
    """get bathymetry of the Northsea from the jarkus dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data where bathymetry is added to
    gridprops : dic, optional
        model properties when using vertex grids. The default is None.

    Returns
    -------
    model_ds_out : xarray.Dataset
        dataset with bathymetry

    Notes
    -----
    The nan values in the original bathymetry are filled and then the
    data is resampled to the modelgrid. Maybe we can speed up things by
    changing the order in which operations are executed.
    """
    model_ds_out = util.get_model_ds_empty(model_ds)

    # no bathymetry if we don't have northsea
    if (northsea == 0).all():
        model_ds_out['bathymetry'] = util.get_da_from_da_ds(northsea,
                                                            northsea.dims,
                                                            data=np.nan)
        return model_ds_out

    # try to get bathymetry via opendap
    try:
        url = 'https://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/grids/catalog.nc'
        jarkus_ds = get_dataset_jarkus(model_ds.extent, url)
    except OSError:
        import gdown
        logger.warning(
            'cannot access Jarkus netCDF link, copy file from google drive instead')
        fname_jarkus = os.path.join(model_ds.model_ws,
                                    'jarkus_nhflopy.nc')
        url = 'https://drive.google.com/uc?id=1uNy4THL3FmNFrTDTfizDAl0lxOH-yCEo'
        gdown.download(url, fname_jarkus,
                       quiet=False)
        jarkus_ds = xr.open_dataset(fname_jarkus)

    da_bathymetry_raw = jarkus_ds['z']

    # fill nan values in bathymetry
    da_bathymetry_filled = mdims.fillnan_dataarray_structured_grid(
        da_bathymetry_raw)

    # bathymetrie mag nooit groter zijn dan NAP 0.0
    da_bathymetry_filled = xr.where(
        da_bathymetry_filled > 0, 0, da_bathymetry_filled)

    # bathymetry projected on model grid
    if model_ds.gridtype == 'structured':
        da_bathymetry = mdims.resample_dataarray2d_to_structured_grid(da_bathymetry_filled,
                                                                      extent=model_ds.extent,
                                                                      delr=model_ds.delr,
                                                                      delc=model_ds.delc,
                                                                      xmid=model_ds.x.data,
                                                                      ymid=model_ds.y.data)
    elif model_ds.gridtype == 'vertex':
        da_bathymetry = mdims.resample_dataarray2d_to_vertex_grid(da_bathymetry_filled,
                                                                  gridprops=gridprops)

    model_ds_out['bathymetry'] = xr.where(northsea, da_bathymetry, np.nan)

    for datavar in model_ds_out:
        model_ds_out[datavar].attrs['source'] = 'Jarkus'
        model_ds_out[datavar].attrs['url'] = url
        model_ds_out[datavar].attrs['source'] = dt.datetime.now().strftime(
            '%Y%m%d')
        if datavar == 'bathymetry':
            model_ds_out[datavar].attrs['units'] = 'mNAP'

    return model_ds_out


def get_dataset_jarkus(extent, url='http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/grids/catalog.nc'):
    """Get bathymetry from Jarkus within a certain extent. The following steps
    are used:

       1. find Jarkus tiles within the extent
       2. combine netcdf urls of Jarkus tiles
       3. read Jarkus tiles and combine the 'z' parameter of the last time step
          of each tile, to a dataarray.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax) of the desired grid. Should be RD-new
        coördinates (EPSG:28992)

    Returns
    -------
    z : xr.DataSet
        dataset containing bathymetry data
    """

    extent = [int(x) for x in extent]
    netcdf_tile_names = get_jarkus_tilenames(extent, url=url)
    tiles = [xr.open_dataset(name) for name in netcdf_tile_names]
    # only use the last timestep
    tiles = [tile.isel(time=-1) for tile in tiles]
    z_dataset = xr.combine_by_coords(tiles, combine_attrs='drop')

    return z_dataset


def get_jarkus_tilenames(extent, url='http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/grids/catalog.nc'):
    """Find all Jarkus tilenames within a certain extent.

    Parameters
    ----------
    extent : list, tuple or np.array
        extent (xmin, xmax, ymin, ymax) of the desired grid. Should be RD-new
        coördinates (EPSG:28992)

    Returns
    -------
    netcdf_urls : list of str
        list of the urls of all netcdf files of the tiles with Jarkus data.
    """
    ds_jarkus_catalog = xr.open_dataset(url)
    ew_x = ds_jarkus_catalog['projectionCoverage_x']
    sn_y = ds_jarkus_catalog['projectionCoverage_y']

    mask_ew = (ew_x[:, 1] > extent[0]) & (ew_x[:, 0] < extent[1])
    mask_sn = (sn_y[:, 1] > extent[2]) & (sn_y[:, 0] < extent[3])

    indices_tiles = np.where(mask_ew & mask_sn)[0]
    all_netcdf_tilenames = get_netcdf_tiles()

    netcdf_tile_names = [all_netcdf_tilenames[i] for i in indices_tiles]

    return netcdf_tile_names


def get_netcdf_tiles():
    """Find all Jarkus netcdf tile names.

    Returns
    -------
    netcdf_urls : list of str
        list of the urls of all netcdf files of the tiles with Jarkus data.

    Notes
    -----
    This function would be redundant if the jarkus catalog
    (http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/grids/catalog.nc)
    had a proper way of displaying the url's of each tile. It seems like an
    attempt was made to do this because there is a data variable
    named 'urlPath' in the catalog. However the dataarray of 'urlPath' has the
    same string for each tile.
    """
    url = 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/grids/catalog.nc.ascii'
    req = requests.get(url)
    s = req.content.decode('ascii')
    start = s.find('urlPath', s.find('urlPath') + 1)
    end = s.find('projectionCoverage_x', s.find('projectionCoverage_x') + 1)
    netcdf_urls = list(eval(s[start + 12:end - 2]))
    return netcdf_urls


def add_bathymetry_to_top_bot_kh_kv(model_ds, bathymetry,
                                    fill_mask,
                                    kh_sea=10,
                                    kv_sea=10):
    """add bathymetry to the top and bot of each layer for all cells with
    fill_mask.

    Parameters
    ----------
    model_ds : xarray.Dataset
        dataset with model data, should
    bathymetry : xarray DataArray
        bathymetry data
    kh_sea : int or float, optional
        the horizontal conductance in sea s
    fill_mask : xr.DataArray
        cell value is 1 if you want to add bathymetry

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with model data where the top, bot, kh and kv are changed
    """
    model_ds['top'].values = np.where(fill_mask,
                                      0.0,
                                      model_ds['top'])

    lay = 0
    model_ds['bot'][lay] = xr.where(fill_mask,
                                    bathymetry,
                                    model_ds['bot'][lay])

    model_ds['kh'][lay] = xr.where(fill_mask,
                                   kh_sea,
                                   model_ds['kh'][lay])

    model_ds['kv'][lay] = xr.where(fill_mask,
                                   kv_sea,
                                   model_ds['kv'][lay])

    # reset bot for all layers based on bathymetrie
    for lay in range(1, model_ds.dims['layer']):
        model_ds['bot'][lay] = np.where(model_ds['bot'][lay] > model_ds['bot'][lay - 1],
                                        model_ds['bot'][lay - 1],
                                        model_ds['bot'][lay])

    return model_ds



# -*- coding: utf-8 -*-
"""Created on Mon Nov 29 12:45:23 2021.

@author: oebbe
"""
import functools
import importlib
import inspect
import logging
import numbers
import os
import pickle

import flopy
import numpy as np
import pandas as pd
import xarray as xr

logger = logging.getLogger(__name__)


def clear_cache(cachedir):
    """clears the cache in a given cache directory by removing all .pklz and
    corresponding .nc files.

    Parameters
    ----------
    cachedir : str
        path to cache directory.

    Returns
    -------
    None.
    """
    ans = input(
        f'this will remove all cached files in {cachedir} are you sure [Y/N]')
    if ans.lower() != 'y':
        return

    for fname in os.listdir(cachedir):
        # assuming all pklz files belong to a cached netcdf file
        if fname.endswith('.pklz'):
            fname_nc = fname.replace('.pklz', '.nc')

            # remove pklz file
            os.remove(os.path.join(cachedir, fname))

            # remove netcdf file (make sure cached netcdf is closed)
            fpath_nc = os.path.join(cachedir, fname_nc)
            cached_ds = xr.open_dataset(fpath_nc)
            cached_ds.close()

            os.remove(fpath_nc)
            logger.info(f'removing {fname} and {fname_nc}')


def cache_netcdf(func):
    """decorator to read/write the result of a function from/to a file to speed
    up function calls with the same arguments. Should only be applied to
    functions that:

        - return an Xarray Dataset
        - have no more than one xarray dataset as function argument
        - have functions arguments of types that can be checked using the
        _is_valid_cache functions

    1. The directory and filename of the cache should be defined by the person
    calling a function with this decorator. If not defined no cache is
    created nor used.
    2. Create a new cached file if it is impossible to check if the function
    arguments used to create the cached file are the same as the current
    function arguments. This can happen if one of the function arguments has a
    type that cannot be checked using the _is_valid_cache function.
    3. Function arguments are pickled together with the cache to check later
    if the cache is valid.
    4. If one of the function arguments is an xarray Dataset it is not pickled.
    Therefore we cannot check if this function argument is identical for the
    cached data and the new function call. We do check if the xarray Dataset
    coördinates correspond to the coördinates of the cached netcdf file.
    5. This function uses `functools.wraps` and some home made
    magic in _update_docstring_and_signature to add arguments of the decorator
    to the decorated function. This assumes that the decorated function has a
    docstring with a "Returns" heading. If this is not the case an error is
    raised when trying to decorate the function.
    """

    # add cachedir and cachename to docstring
    _update_docstring_and_signature(func)

    @functools.wraps(func)
    def decorator(*args, cachedir=None, cachename=None, **kwargs):

        if cachedir is None or cachename is None:
            return func(*args, **kwargs)

        if not cachename.endswith('.nc'):
            cachename += '.nc'

        fname_cache = os.path.join(cachedir, cachename)  # netcdf file
        fname_pickle_cache = fname_cache.replace(
            '.nc', '.pklz')  # pickle with function arguments

        # create dictionary with function arguments
        func_args_dic = {f'arg{i}': args[i] for i in range(len(args))}
        func_args_dic.update(kwargs)

        # remove xarray dataset from function arguments
        dataset = None
        for key in list(func_args_dic.keys()):
            if isinstance(func_args_dic[key], xr.Dataset):
                if dataset is not None:
                    raise TypeError(
                        'function was called with multiple xarray dataset arguments')
                dataset = func_args_dic.pop(key)

        # only use cache if the cache file and the pickled function arguments exist
        if os.path.exists(fname_cache) and os.path.exists(fname_pickle_cache):
            with open(fname_pickle_cache, "rb") as f:
                func_args_dic_cache = pickle.load(f)

            # check if the module where the function is defined was changed
            # after the cache was created
            time_mod_func = _get_modification_time(func)
            time_mod_cache = os.path.getmtime(fname_cache)
            modification_check = time_mod_cache > time_mod_func

            if not modification_check:
                logger.info(
                    f'module of function {func.__name__} recently modified, not using cache')

            # check if cache was created with same function arguments as
            # function call
            argument_check = _same_function_arguments(func_args_dic,
                                                      func_args_dic_cache)

            if modification_check and argument_check:
                cached_ds = xr.open_dataset(fname_cache)
                if dataset is None:
                    logger.info(f'using cached data -> {cachename}')
                    return cached_ds

                # check if cached dataset has same dimension and coordinates
                # as current dataset
                if _check_ds(dataset, cached_ds):
                    logger.info(f'using cached data -> {cachename}')
                    return cached_ds

        # create cache
        result = func(*args, **kwargs)
        logger.info(f'caching data -> {cachename}')

        if isinstance(result, xr.Dataset):
            # close cached netcdf (otherwise it is impossible to overwrite)
            if os.path.exists(fname_cache):
                cached_ds = xr.open_dataset(fname_cache)
                cached_ds.close()

            # write netcdf cache
            result.to_netcdf(fname_cache)
            # pickle function arguments
            with open(fname_pickle_cache, 'wb') as fpklz:
                pickle.dump(func_args_dic, fpklz)
        else:
            raise TypeError(
                f'expected xarray Dataset, got {type(result)} instead')

        return result

    return decorator


def cache_pklz(func):
    """decorator to read/write the result of a function from/to a pklz file to
    speed up function calls with the same arguments. Should only be applied to
    functions that:

        - return a dictionary
        - have functions arguments of types that can be checked using the
        _is_valid_cache functions

    1. The directory and filename of the cache should be defined by the person
    calling a function with this decorator. If not defined no cache is
    created nor used.
    2. Create a new cached file if it is impossible to check if the function
    arguments used to create the cached file are the same as the current
    function arguments. This can happen if one of the function arguments has a
    type that cannot be checked using the _is_valid_cache function.
    3. Function arguments are pickled together with the cache to check later
    if the cache is valid.
    4. This function uses `functools.wraps` and some home made
    magic in _update_docstring_and_signature to add arguments of the decorator
    to the decorated function. This assumes that the decorated function has a
    docstring with a "Returns" heading. If this is not the case an error is
    raised when trying to decorate the function.
    """

    _update_docstring_and_signature(func)

    @functools.wraps(func)
    def decorator(*args, cachedir=None, cachename=None, **kwargs):

        if cachedir is None or cachename is None:
            return func(*args, **kwargs)

        if not cachename.endswith('.pklz'):
            cachename += '.pklz'

        fname_cache = os.path.join(cachedir, cachename)  # pklz file
        fname_args_cache = fname_cache.replace(
            '.pklz', '_cache.pklz')  # pickle with function arguments

        # create dictionary with function arguments
        func_args_dic = {f'arg{i}': args[i] for i in range(len(args))}
        func_args_dic.update(kwargs)

        # only use cache if the cache file and the pickled function arguments exist
        if os.path.exists(fname_cache) and os.path.exists(fname_args_cache):
            with open(fname_args_cache, "rb") as f:
                func_args_dic_cache = pickle.load(f)

            # check if the module where the function is defined was changed
            # after the cache was created
            time_mod_func = _get_modification_time(func)
            time_mod_cache = os.path.getmtime(fname_cache)
            modification_check = time_mod_cache > time_mod_func

            if not modification_check:
                logger.info(
                    f'module of function {func.__name__} recently modified, not using cache')

            # check if cache was created with same function arguments as
            # function call
            argument_check = _same_function_arguments(func_args_dic,
                                                      func_args_dic_cache)

            if modification_check and argument_check:
                with open(fname_cache, "rb") as f:
                    result = pickle.load(f)
                logger.info(f'using cached data -> {cachename}')
                return result

        # create cache
        result = func(*args, **kwargs)
        logger.info(f'caching data -> {cachename}')

        if isinstance(result, dict):
            # write result
            with open(fname_cache, "wb") as f:
                pickle.dump(result, f)
            # pickle function arguments
            with open(fname_args_cache, 'wb') as fpklz:
                pickle.dump(func_args_dic, fpklz)
        else:
            raise TypeError(f'expected dictionary, got {type(result)} instead')

        return result

    return decorator


def _check_ds(ds, ds2):
    """check if two datasets have the same dimensions and coordinates.

    Parameters
    ----------
    ds : xr.Dataset
        dataset with dimensions and coordinates
    ds2 : xr.Dataset
        dataset with dimensions and coordinates. This is typically
        a cached dataset.

    Returns
    -------
    bool
        True if the two datasets have the same grid and time discretization.
    """
    for coord in ds2.coords:
        if coord in ds.coords:
            if not ds2[coord].equals(ds[coord]):
                logger.info(
                    f'coordinate {coord} has different values in cached dataset, not using cache')
                return False
        else:
            logger.info(
                f'dimension {coord} only present in cache, not using cache')
            return False

    return True


def _same_function_arguments(func_args_dic, func_args_dic_cache):
    """ checks if two dictionaries with function arguments are identical by
    checking:
        1. if they have the same keys
        2. if the items have the same type
        3. if the items have the same values (only possible for the types: int,
                                              float, bool, str, bytes, list,
                                              tuple, dict, np.ndarray,
                                              xr.DataArray,
                                              flopy.mf6.ModflowGwf)

    Parameters
    ----------
    func_args_dic : dictionary
        dictionary with all the args and kwargs of a function call.
    func_args_dic_cache : dictionary
        dictionary with all the args and kwargs of a previous function call of
        which the results are cached.

    Returns
    -------
    bool
        if True the dictionaries are identical which means that the cached
        data was created using the same function arguments as the requested
        data.

    """
    for key, item in func_args_dic.items():
        # check if cache and function call have same argument names
        if key not in func_args_dic_cache.keys():
            logger.info(
                'cache was created using different function arguments, do not use cached data')
            return False

        # check if cache and function call have same argument types
        if not isinstance(item, type(func_args_dic_cache[key])):
            logger.info(
                'cache was created using different function argument types, do not use cached data')
            return False

        # check if cache and function call have same argument values
        if item is None:
            # Value of None type is always None so the check happens in previous if statement
            pass
        elif isinstance(item, (numbers.Number, bool, str, bytes, list, tuple)):
            if item != func_args_dic_cache[key]:
                logger.info(
                    'cache was created using different function argument values, do not use cached data')
                return False
        elif isinstance(item, np.ndarray):
            if not np.allclose(item, func_args_dic_cache[key]):
                logger.info(
                    'cache was created using different numpy array values, do not use cached data')
                return False
        elif isinstance(item, (pd.DataFrame, pd.Series, xr.DataArray)):
            if not item.equals(func_args_dic_cache[key]):
                logger.info(
                    'cache was created using different DataFrame/Series/DataArray, do not use cached data')
                return False
        elif isinstance(item, dict):
            # recursive checking
            if not _same_function_arguments(item, func_args_dic_cache[key]):
                logger.info(
                    'cache was created using different dictionaries, do not use cached data')
                return False
        elif isinstance(item, (flopy.mf6.ModflowGwf, flopy.modflow.mf.Modflow)):
            if str(item) != str(func_args_dic_cache[key]):
                logger.info(
                    'cache was created using different groundwater flow model, do not use cached data')
                return False

        else:
            logger.info(
                'cannot check if cache is valid, assuming invalid cache')
            logger.info(f'function argument of type {type(item)}')
            return False

    return True


def _get_modification_time(func):
    """return the modification time of the module where func is defined.

    Parameters
    ----------
    func : function
        function.

    Returns
    -------
    float
        modification time of module.
    """
    mod = func.__module__
    active_mod = importlib.import_module(mod.split('.')[0])
    if '.' in mod:
        for submod in mod.split('.')[1:]:
            active_mod = getattr(active_mod, submod)

    return os.path.getmtime(active_mod.__file__)


def _update_docstring_and_signature(func):
    """add function arguments 'cachedir' and 'cachename' to the docstring and
    signature of a function. The function arguments are added before the
    "Returns" header in the docstring. If the function has no Returns header in
    the docstring, the function arguments are not added to the docstring.

    Parameters
    ----------
    func : function
        function that is decorated.

    Returns
    -------
    None.
    """
    # add cachedir and cachename to signature
    sig = inspect.signature(func)
    cur_param = tuple(sig.parameters.values())
    new_param = cur_param + (inspect.Parameter('cachedir',
                                               inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                               default=None),
                             inspect.Parameter('cachename',
                                               inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                               default=None))
    sig = sig.replace(parameters=new_param)
    func.__signature__ = sig

    # add cachedir and cachename to docstring
    original_doc = func.__doc__
    if original_doc is None:
        logger.warning(f'Function "{func.__name__}" has no docstring')
        return
    if 'Returns' not in original_doc:
        logger.warning(
            f'Function "{func.__name__}" has no "Returns" header in docstring')
        return
    before, after = original_doc.split('Returns')
    mod_before = before.strip() + '\n    cachedir : str or None, optional\n'\
        '        directory to save cache. If None no cache is used.'\
        ' Default is None.\n    cachename : str or None, optional\n'\
        '        filename of netcdf cache. If None no cache is used.'\
        ' Default is None.\n\n    Returns'
    new_doc = ''.join((mod_before, after))
    func.__doc__ = new_doc
    return



import datetime as dt
import os
import sys

import xarray as xr

from .. import util


def get_empty_model_ds(model_name, model_ws, mfversion="mf6",
                       exe_name=None):
    """get an empty model dataset.

    Parameters
    ----------
    model_name : str
        name of the model.
    model_ws : str or None
        workspace of the model. This is where modeldata is saved to.
    mfversion : str, optional
        modflow version. The default is "mf6".
    exe_name: str, optional
        path to modflow executable, default is None, which assumes binaries
        are available in nlmod/bin directory. Binaries can be downloaded
        using `nlmod.util.download_mfbinaries()`.

    Returns
    -------
    model_ds : xarray dataset
        model dataset.
    """

    model_ds = xr.Dataset()

    model_ds.attrs['model_name'] = model_name
    model_ds.attrs['mfversion'] = mfversion
    model_ds.attrs['model_dataset_created_on'] = dt.datetime.now().strftime(
        "%Y%m%d_%H:%M:%S")

    if exe_name is None:
        exe_name = os.path.join(os.path.dirname(__file__),
                                '..', '..', 'bin', model_ds.mfversion)

    # if working on Windows add .exe extension
    if sys.platform.startswith('win'):
        exe_name += ".exe"

    model_ds.attrs["exe_name"] = exe_name

    # add some directories
    if model_ws is not None:
        figdir, cachedir = util.get_model_dirs(model_ws)
        model_ds.attrs['model_ws'] = model_ws
        model_ds.attrs['figdir'] = figdir
        model_ds.attrs['cachedir'] = cachedir

    return model_ds



# -*- coding: utf-8 -*-
"""functions to add surface water to a mf model using the ghb package."""

import datetime as dt
import logging
import os

import geopandas as gpd
import nlmod
import xarray as xr

from .. import cache, mdims, util

logger = logging.getLogger(__name__)


def get_gdf_surface_water(model_ds):
    """read a shapefile with surface water as a geodataframe, cut by the extent
    of the model.

    Parameters
    ----------
    model_ds : xr.DataSet
        dataset containing relevant model information

    Returns
    -------
    gdf_opp_water : GeoDataframe
        surface water geodataframe.
    """
    # laad bestanden in
    fname = os.path.join(nlmod.NLMOD_DATADIR, 'opp_water.shp')
    gdf_swater = gpd.read_file(fname)
    gdf_swater = util.gdf_within_extent(gdf_swater, model_ds.extent)

    return gdf_swater


@cache.cache_netcdf
def get_surface_water(model_ds, da_name, gridprops=None):
    """create 3 data-arrays from the shapefile with surface water:

    - area: with the area of the shape in the cell
    - cond: with the conductance based on the area and bweerstand column in shapefile
    - peil: with the surface water lvl based on the peil column in the shapefile

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    da_name : str
        name of the polygon shapes, name is used to store data arrays in
        model_ds
    gridprops : dict, optional
        extra model properties when using unstructured grids.
        The default is None.

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with modelgrid data.
    """

    modelgrid = mdims.modelgrid_from_model_ds(model_ds, gridprops=gridprops)
    gdf = get_gdf_surface_water(model_ds)

    area = xr.zeros_like(model_ds['top'])
    cond = xr.zeros_like(model_ds['top'])
    peil = xr.zeros_like(model_ds['top'])
    for _, row in gdf.iterrows():
        area_pol = mdims.polygon_to_area(modelgrid, row['geometry'],
                                         xr.ones_like(model_ds['top']),
                                         model_ds.gridtype)
        cond = xr.where(area_pol > area, area_pol / row['bweerstand'], cond)
        peil = xr.where(area_pol > area, row['peil'], peil)
        area = xr.where(area_pol > area, area_pol, area)

    model_ds_out = util.get_model_ds_empty(model_ds)
    model_ds_out[f'{da_name}_area'] = area
    model_ds_out[f'{da_name}_area'].attrs['units'] = 'm2'
    model_ds_out[f'{da_name}_cond'] = cond
    model_ds_out[f'{da_name}_cond'].attrs['units'] = 'm2/day'
    model_ds_out[f'{da_name}_peil'] = peil
    model_ds_out[f'{da_name}_peil'].attrs['units'] = 'mNAP'

    for datavar in model_ds_out:
        model_ds_out[datavar].attrs['source'] = 'RWS'
        model_ds_out[datavar].attrs['date'] = dt.datetime.now().strftime(
            '%Y%m%d')

    return model_ds_out


@cache.cache_netcdf
def get_northsea(model_ds, gridprops=None, da_name='northsea'):
    """Get Dataset which is 1 at the northsea and 0 everywhere else. Sea is
    defined by rws surface water shapefile.

    Parameters
    ----------
    model_ds : xr.DataSet
        xarray with model data
    gridprops : dictionary
        dictionary with grid properties output from gridgen.
    da_name : str, optional
        name of the datavar that identifies sea cells

    Returns
    -------
    model_ds_out : xr.DataSet
        Dataset with a single DataArray, this DataArray is 1 at sea and 0
        everywhere else. Grid dimensions according to model_ds.
    """

    gdf_surf_water = get_gdf_surface_water(model_ds)

    # find grid cells with sea
    swater_zee = gdf_surf_water[gdf_surf_water['OWMNAAM'].isin(['Rijn territoriaal water',
                                                                'Waddenzee',
                                                                'Waddenzee vastelandskust',
                                                                'Hollandse kust (kustwater)',
                                                                'Waddenkust (kustwater)'])]

    modelgrid = mdims.modelgrid_from_model_ds(model_ds, gridprops=gridprops)
    model_ds_out = mdims.gdf_to_bool_dataset(model_ds, swater_zee,
                                             modelgrid, da_name)

    return model_ds_out



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    l_d = f.read()

# Get the version.
version = {}
with open("nlmod/version.py") as fp:
    exec(fp.read(), version)

setup(
    name='nlmod',
    version=version['__version__'],
    description='nlmod module by Artesia',
    long_description=l_d,
    long_description_content_type="text/markdown",
    url='https://github.com/ArtesiaWater/nlmod',
    author='Artesia',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    platforms='Windows, Mac OS-X',
    install_requires=['flopy>=3.3.2',
                      'xarray>=0.16.1',
					  'rasterio>=1.1.0',
					  'owslib>=0.24.1',
					  'hydropandas>=0.3.0',
					  'netcdf4>=1.5.7',
					  'pyshp>=2.1.3',
					  'rtree>=0.9.7',
					  'openpyxl>=3.0.7',
                      'matplotlib'
                      ],
    packages=find_packages(exclude=[]),
    package_data={"nlmod": ["data/*"]},
    include_package_data=True
)



# -*- coding: utf-8 -*-
"""add knmi precipitation and evaporation to a modflow model."""


import logging

import flopy
import numpy as np
import xarray as xr
from tqdm import tqdm

from .. import mdims
from . import mfpackages

logger = logging.getLogger(__name__)


def model_datasets_to_rch(gwf, model_ds, print_input=False):
    """convert the recharge data in the model dataset to a recharge package
    with time series.

    Parameters
    ----------
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        groundwater flow model.
    model_ds : xr.DataSet
        dataset containing relevant model grid information
    print_input : bool, optional
        value is passed to flopy.mf6.ModflowGwfrch() to determine if input
        should be printed to the lst file. Default is False

    Returns
    -------
    rch : flopy.mf6.modflow.mfgwfrch.ModflowGwfrch
        recharge package
    """
    # check for nan values
    if model_ds['recharge'].isnull().any():
        raise ValueError('please remove nan values in recharge data array')

    # get stress period data
    if model_ds.steady_state:
        mask = model_ds['recharge'] != 0
        if model_ds.gridtype == 'structured':
            rch_spd_data = mdims.data_array_2d_to_rec_list(
                model_ds, mask, col1='recharge',
                first_active_layer=True,
                only_active_cells=False)
        elif model_ds.gridtype == 'vertex':
            rch_spd_data = mdims.data_array_1d_vertex_to_rec_list(
                model_ds, mask, col1='recharge',
                first_active_layer=True,
                only_active_cells=False)

        # create rch package
        rch = flopy.mf6.ModflowGwfrch(gwf,
                                      filename=f'{gwf.name}.rch',
                                      pname=f'{gwf.name}',
                                      fixed_cell=False,
                                      maxbound=len(rch_spd_data),
                                      print_input=True,
                                      stress_period_data={0: rch_spd_data})

        return rch

    # transient recharge
    if model_ds.gridtype == 'structured':
        empty_str_array = np.zeros_like(model_ds['idomain'][0], dtype="S13")
        model_ds['rch_name'] = xr.DataArray(empty_str_array,
                                            dims=('y', 'x'),
                                            coords={'y': model_ds.y,
                                                    'x': model_ds.x})
        model_ds['rch_name'] = model_ds['rch_name'].astype(str)
        # dimension check
        if model_ds['recharge'].dims == ('time', 'y', 'x'):
            axis = 0
            rch_2d_arr = model_ds['recharge'].data.reshape(
                (model_ds.dims['time'], model_ds.dims['x'] * model_ds.dims['y'])).T

            # check if reshaping is correct
            if not (model_ds['recharge'].values[:, 0, 0] == rch_2d_arr[0]).all():
                raise ValueError(
                    'reshaping recharge to calculate unique time series did not work out as expected')

        elif model_ds['recharge'].dims == ('y', 'x', 'time'):
            axis = 2
            rch_2d_arr = model_ds['recharge'].data.reshape(
                (model_ds.dims['x'] * model_ds.dims['y'], model_ds.dims['time']))

            # check if reshaping is correct
            if not (model_ds['recharge'].values[0, 0, :] == rch_2d_arr[0]).all():
                raise ValueError(
                    'reshaping recharge to calculate unique time series did not work out as expected')

        else:
            raise ValueError('expected dataarray with 3 dimensions'
                             f'(time, y and x) or (y, x and time), not {model_ds["recharge"].dims}')

        rch_unique_arr = np.unique(rch_2d_arr, axis=0)
        rch_unique_dic = {}
        for i, unique_rch in enumerate(rch_unique_arr):
            model_ds['rch_name'].data[np.isin(
                model_ds['recharge'].values, unique_rch).all(axis=axis)] = f'rch_{i}'
            rch_unique_dic[f'rch_{i}'] = unique_rch

        mask = model_ds['rch_name'] != ''
        rch_spd_data = mdims.data_array_2d_to_rec_list(model_ds, mask,
                                                       col1='rch_name',
                                                       first_active_layer=True,
                                                       only_active_cells=False)

    elif model_ds.gridtype == 'vertex':
        empty_str_array = np.zeros_like(model_ds['idomain'][0], dtype="S13")
        model_ds['rch_name'] = xr.DataArray(empty_str_array,
                                            dims=('cid'),
                                            coords={'cid': model_ds.cid})
        model_ds['rch_name'] = model_ds['rch_name'].astype(str)

        # dimension check
        if model_ds['recharge'].dims == ('cid', 'time'):
            rch_2d_arr = model_ds['recharge'].values
        elif model_ds['recharge'].dims == ('time', 'cid'):
            rch_2d_arr = model_ds['recharge'].values.T
        else:
            raise ValueError('expected dataarray with 2 dimensions'
                             f'(time, cid) or (cid, time), not {model_ds["recharge"].dims}')

        rch_unique_arr = np.unique(rch_2d_arr, axis=0)
        rch_unique_dic = {}
        for i, unique_rch in enumerate(rch_unique_arr):
            model_ds['rch_name'][(rch_2d_arr == unique_rch).all(
                axis=1)] = f'rch_{i}'
            rch_unique_dic[f'rch_{i}'] = unique_rch

        mask = model_ds['rch_name'] != ''
        rch_spd_data = mdims.data_array_1d_vertex_to_rec_list(model_ds, mask,
                                                              col1='rch_name',
                                                              first_active_layer=True,
                                                              only_active_cells=False)

    # create rch package
    rch = flopy.mf6.ModflowGwfrch(gwf, filename=f'{gwf.name}.rch',
                                  pname='rch',
                                  fixed_cell=False,
                                  maxbound=len(rch_spd_data),
                                  print_input=print_input,
                                  stress_period_data={0: rch_spd_data})

    # get timesteps
    tdis_perioddata = mfpackages.get_tdis_perioddata(model_ds)
    perlen_arr = [t[0] for t in tdis_perioddata]
    time_steps_rch = [0.0] + np.array(perlen_arr).cumsum().tolist()

    # create timeseries packages
    for i, key in tqdm(enumerate(rch_unique_dic.keys()),
                       total=len(rch_unique_dic.keys()),
                       desc="Building ts packages rch"):
        # add extra time step to the time series object (otherwise flopy fails)
        recharge_val = list(rch_unique_dic[key]) + [0.0]

        recharge = list(zip(time_steps_rch, recharge_val))
        if i == 0:
            rch.ts.initialize(filename=f'{key}.ts',
                              timeseries=recharge,
                              time_series_namerecord=key,
                              interpolation_methodrecord='stepwise')
        else:
            rch.ts.append_package(filename=f'{key}.ts',
                                  timeseries=recharge,
                                  time_series_namerecord=key,
                                  interpolation_methodrecord='stepwise')

    return rch



# -*- coding: utf-8 -*-
"""Created on Fri Apr 17 13:50:48 2020.

@author: oebbe
"""

import datetime as dt
import logging

import numpy as np
import pandas as pd
import xarray as xr

logger = logging.getLogger(__name__)


def set_model_ds_time(model_ds, start_time, steady_state,
                      steady_start=False, time_units='DAYS',
                      transient_timesteps=0,
                      perlen=1.0, steady_perlen=3650,
                      nstp=1, tsmult=1.0):
    """Set timing for a model dataset.

    Parameters
    ----------
    model_ds : xarray.Dataset
        Dataset to add time information to
    start_time : str or datetime
        start time of the model. This is the start_time of the transient time
        steps.
    steady_state : bool
        if True the model is steady state with one time step.
    steady_start : bool
        if True the model is transient with a steady state start time step.
    time_units : str, optional
        time unit of the model. The default is 'DAYS'.
    transient_timesteps : int, optional
        number of transient time steps. The default is 0.
    perlen : float, int, list or np.array, optional
        length of each timestep depending on the type, the default is 1.0:
        - float or int: this is the length of all the time steps. If
          steady_start is True the length of the first time step is defined
          by steady_perlen
        - list or array: the items are the length per timestep.
          the length of perlen should match the number of transient
          timesteps (or transient timesteps +1 if steady_start is True)
    steady_perlen : float, optional
        time step length of the first steady state timestep.
        Only used if steady_start is True. Default is 3650 (~10 years)
    nstp : int, optional
        number of steps. The default is 1.
    tsmult : float, optional
        timestep multiplier. The default is 1.0.

    Returns
    -------
    model_ds : xarray.Dataset
        dataset with time variant model data
    """
    # checks
    if len(model_ds.model_name) > 16 and model_ds.mfversion == 'mf6':
        raise ValueError('model_name can not have more than 16 characters')
    elif time_units.lower() != 'days':
        raise NotImplementedError()
    elif (not steady_state) and (steady_start and (transient_timesteps == 0)):
        raise ValueError('illegal combination of steady_start and transient'
                         'timesteps please use steady_state=True')

    start_time_dt = pd.to_datetime(start_time)

    if steady_state:
        nper = 1
        start_time_dt = pd.to_datetime(start_time)
        time_dt = [start_time_dt]
        if perlen == 1.0:
            perlen = steady_perlen
            logger.warning(f'setting perlen to {steady_perlen}')
    elif steady_start:
        nper = 1 + transient_timesteps

        if isinstance(perlen, (float, int)):
            start_tran = pd.to_datetime(start_time)
            start_time_dt = start_tran - dt.timedelta(days=steady_perlen)

            time_dt = pd.date_range(start_tran - pd.to_timedelta(perlen, unit=time_units),
                                    start_tran + pd.to_timedelta((transient_timesteps - 1) * perlen,
                                                                 unit=time_units), periods=nper)
            time_dt.values[0] = start_time_dt

        elif isinstance(perlen, (list, np.ndarray)):
            assert len(perlen) == nper
            start_tran = pd.to_datetime(start_time)
            start_time_dt = start_tran - dt.timedelta(days=perlen[0])

            time_dt = [start_time_dt]
            for i, p in enumerate(perlen[:-1]):
                time_dt += [time_dt[i] + pd.to_timedelta(p, unit=time_units)]
        else:
            raise ValueError(f'did not recognise perlen data type {perlen}')

    else:
        nper = transient_timesteps
        if isinstance(perlen, (float, int)):
            start_time_dt = pd.to_datetime(start_time)
            time_dt = pd.date_range(start_time_dt,
                                    start_time_dt + pd.to_timedelta((transient_timesteps - 1) * perlen,
                                                                    unit=time_units), periods=nper)
        elif isinstance(perlen, (list, np.ndarray)):
            assert len(perlen) == nper
            start_time_dt = pd.to_datetime(start_time)
            time_dt = [start_time_dt]
            for i, p in enumerate(perlen[:-1]):
                time_dt += [time_dt[i] + pd.to_timedelta(p, unit=time_units)]
        else:
            raise ValueError(f'did not recognise perlen data type {perlen}')

    time_steps = list(range(nper))

    model_ds = model_ds.assign_coords(coords={'time': time_dt})

    model_ds['time_steps'] = xr.DataArray(time_steps, dims=('time'),
                                          coords={'time': model_ds.time})
    model_ds.attrs['time_units'] = time_units
    model_ds.attrs['start_time'] = str(start_time_dt)
    model_ds.attrs['nper'] = nper
    model_ds.attrs['perlen'] = perlen
    model_ds.attrs['nstp'] = nstp
    model_ds.attrs['tsmult'] = tsmult

    # netcdf files cannot handle booleans
    model_ds.attrs['steady_start'] = int(steady_start)
    model_ds.attrs['steady_state'] = int(steady_state)

    return model_ds



# -*- coding: utf-8 -*-
"""Created on Thu Jan  7 21:32:49 2021.

@author: oebbe
"""

import os
import warnings

import flopy
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

from ..read import rws


def plot_surface_water(model_ds, ax=None):
    surf_water = rws.get_gdf_surface_water(model_ds)

    if ax is None:
        _, ax = plt.subplots()
    surf_water.plot(ax=ax)

    return ax


def plot_modelgrid(model_ds, gwf, ax=None, add_surface_water=True):

    if ax is None:
        _, ax = plt.subplots(figsize=(10, 10))

    gwf.modelgrid.plot(ax=ax)
    ax.axis('scaled')
    if add_surface_water:
        plot_surface_water(model_ds, ax=ax)
        ax.set_title('modelgrid with surface water')
    else:
        ax.set_title('modelgrid')
    ax.set_ylabel('y [m RD]')
    ax.set_xlabel('x [m RD]')

    return ax


def facet_plot(gwf, arr, lbl="", plot_dim="layer", layer=None, period=None,
               cmap="viridis", scale_cbar=True, vmin=None, vmax=None,
               norm=None, xlim=None, ylim=None, grid=False, figdir=None,
               figsize=(10, 8), plot_bc=None, plot_grid=False):

    if arr.ndim == 4 and plot_dim == "layer":
        nplots = arr.shape[1]
    elif arr.ndim == 4 and plot_dim == "time":
        nplots = arr.shape[0]
    elif arr.ndim == 3:
        nplots = arr.shape[0]
    else:
        raise ValueError("Array must have at least 3 dimensions.")

    plots_per_row = int(np.ceil(np.sqrt(nplots)))
    plots_per_col = nplots // plots_per_row + 1

    fig, axes = plt.subplots(
        plots_per_col, plots_per_row, figsize=figsize,
        sharex=True, sharey=True, constrained_layout=True)

    if scale_cbar:
        vmin = np.nanmin(arr)
        vmax = np.nanmax(arr)

    for i in range(nplots):
        iax = axes.flat[i]
        iax.set_aspect("equal")
        if plot_dim == "layer":
            ilay = i
            iper = period
            if arr.ndim == 4:
                if iper is None:
                    raise ValueError("Pass 'period' to select "
                                     "timestep to plot.")
                a = arr[iper]
        elif plot_dim == "time":
            ilay = layer
            iper = i
            if arr.ndim == 4:
                if ilay is None:
                    raise ValueError("Pass 'layer' to select "
                                     "layer to plot.")
                a = arr[iper]
        else:
            raise ValueError("'plot_dim' must be one of ['layer', 'time']")

        mp = flopy.plot.PlotMapView(model=gwf, layer=ilay, ax=iax)
        qm = mp.plot_array(a, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm)

        mp.plot_ibound(color_vpt="darkgray")

        if plot_grid:
            mp.plot_grid(ls=0.25, color="k")

        if plot_bc is not None:
            for bc, bc_kwargs in plot_bc.items():
                mp.plot_bc(bc, **bc_kwargs)

        iax.grid(grid)
        iax.set_xticklabels([])
        iax.set_yticklabels([])

        if plot_dim == "layer":
            iax.set_title(f"Layer {ilay}", fontsize=6)
        elif plot_dim == "time":
            iax.set_title(f"Timestep {iper}", fontsize=6)

        if xlim is not None:
            iax.set_xlim(xlim)
        if ylim is not None:
            iax.set_ylim(ylim)

    for iax in axes.ravel()[nplots:]:
        iax.set_visible(False)

    cb = fig.colorbar(qm, ax=axes, shrink=1.0)
    cb.set_label(lbl)

    if figdir:
        fig.savefig(os.path.join(figdir, f"{lbl}_per_{plot_dim}.png"),
                    dpi=150, bbox_inches="tight")

    return fig, axes


def facet_plot_ds(gwf, model_ds, figdir, plot_var='bot', plot_time=None,
                  plot_bc=('CHD',), color='k', grid=False,
                  xlim=None, ylim=None):
    """make a 2d plot of every modellayer, store them in a grid.

    Parameters
    ----------
    gwf : Groundwater flow
        Groundwaterflow model.
    model_ds : xr.DataSet
        model data.
    figdir : str
        file path figures.
    plot_var : str, optional
        variable in model_ds. The default is 'bot'.
    plot_time : int, optional
        time step if plot_var is time variant. The default is None.
    plot_bc : list of str, optional
        name of packages of which boundary conditions are plot. The default
        is ['CHD'].
    color : str, optional
        color. The default is 'k'.
    grid : bool, optional
        if True a grid is plotted. The default is False.
    xlim : tuple, optional
        xlimits. The default is None.
    ylim : tuple, optional
        ylimits. The default is None.

    Returns
    -------
    fig : TYPE
        DESCRIPTION.
    axes : TYPE
        DESCRIPTION.
    """
    for key in plot_bc:
        if key not in gwf.get_package_list():
            raise ValueError(
                f'cannot plot boundary condition {key} because it is not in the package list')

    nlay = len(model_ds.layer)

    plots_per_row = int(np.ceil(np.sqrt(nlay)))
    plots_per_col = nlay // plots_per_row + 1

    fig, axes = plt.subplots(
        plots_per_col, plots_per_row, figsize=(11, 10),
        sharex=True, sharey=True, dpi=150
    )
    if plot_time is None:
        plot_arr = model_ds[plot_var]
    else:
        plot_arr = model_ds[plot_var][plot_time]

    vmin = plot_arr.min()
    vmax = plot_arr.max()
    for ilay in range(nlay):
        iax = axes.ravel()[ilay]
        mp = flopy.plot.PlotMapView(model=gwf, layer=ilay, ax=iax)
        # mp.plot_grid()
        qm = mp.plot_array(plot_arr[ilay].values, cmap="viridis",
                           vmin=vmin, vmax=vmax)
        # qm = mp.plot_array(hf[-1], cmap="viridis", vmin=-0.1, vmax=0.1)
        # mp.plot_ibound()
        # plt.colorbar(qm)
        for bc_var in plot_bc:
            mp.plot_bc(bc_var, kper=0, color=color)

        iax.set_aspect("equal", adjustable="box")
        iax.set_title(f"Layer {ilay}")

        iax.grid(grid)
        if xlim is not None:
            iax.set_xlim(xlim)
        if ylim is not None:
            iax.set_ylim(ylim)

    for iax in axes.ravel()[nlay:]:
        iax.set_visible(False)

    cb = fig.colorbar(qm, ax=axes, shrink=1.0)
    cb.set_label(f'{plot_var}', rotation=270)
    fig.suptitle(
        f"{plot_var} Time = {(model_ds.nper*model_ds.perlen)/365} year")
    fig.tight_layout()
    fig.savefig(os.path.join(figdir, f"{plot_var}_per_layer.png"), dpi=150,
                bbox_inches="tight")

    return fig, axes


def plot_array(gwf, array, figsize=(8, 8), colorbar=True, ax=None, **kwargs):

    warnings.warn("The 'plot_array' functions is deprecated please use"
                  "'plot_vertex_array' instead", DeprecationWarning)
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    yticklabels = ax.yaxis.get_ticklabels()
    plt.setp(yticklabels, rotation=90, verticalalignment='center')
    ax.axis('scaled')
    pmv = flopy.plot.PlotMapView(modelgrid=gwf.modelgrid, ax=ax)
    pcm = pmv.plot_array(array, **kwargs)
    if colorbar:
        fig = ax.get_figure()
        fig.colorbar(pcm, ax=ax, orientation='vertical')
        # plt.colorbar(pcm)
    if hasattr(array, 'name'):
        ax.set_title(array.name)
    # set rotation of y ticks to zero
    plt.setp(ax.yaxis.get_majorticklabels(), rotation=0)
    return ax


def plot_vertex_array(da, vertices, ax=None, gridkwargs=None, **kwargs):
    """plot dataarray with gridtype vertex.

    Parameters
    ----------
    da : xarray.Datarray
        plot data with dimension(cid).
    vertices : xarray.Datarray or numpy.ndarray
        Vertex coördinates per cell with dimensions(cid, 4, 2)
    ax : TYPE, optional
        DESCRIPTION. The default is None.
    gridkwargs : dict or None, optional
        layout parameters to plot the cells. For example
        {'edgecolor':'k'} to create black cell lines. The default is None.
    **kwargs :
        passed to quadmesh.set().

    Returns
    -------
    ax : TYPE
        DESCRIPTION.
    """

    if isinstance(vertices, xr.DataArray):
        vertices = vertices.values

    if ax is None:
        _, ax = plt.subplots()

    patches = [Polygon(vert) for vert in vertices]
    if gridkwargs is None:
        quadmesh = PatchCollection(patches)
    else:
        quadmesh = PatchCollection(patches, **gridkwargs)
    quadmesh.set_array(da)

    # set max and min
    if "vmin" in kwargs:
        vmin = kwargs.pop("vmin")
    else:
        vmin = None

    if "vmax" in kwargs:
        vmax = kwargs.pop("vmax")
    else:
        vmax = None

    # limit the color range
    quadmesh.set_clim(vmin=vmin, vmax=vmax)
    quadmesh.set(**kwargs)

    ax.add_collection(quadmesh)
    ax.set_xlim(vertices[:, :, 0].min(), vertices[:, :, 0].max())
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_ylim(vertices[:, :, 1].min(), vertices[:, :, 1].max())
    ax.set_aspect('equal')

    ax.get_figure().colorbar(quadmesh, ax=ax, orientation='vertical')
    if hasattr(da, 'name'):
        ax.set_title(da.name)

    return ax
